/*****************************************************************************/
/*
                                 template.c

Parse and output the template file substituting data for the various sequences.


SUSTITUTIONS
------------
$$$$             escaped double dollar
$$! ... !$$      commented (-out) section
$$ADVISE$$       output an error, warning or informational VWcms message
$$AGO$$          how many minutes/hours/days etc. ago the site was modified
$$BACK$$         JavaScript enabled return to previously viewed page
$$BLOG$$         facilitate weblogging
$$CGI$$          substitute the value of the CGI variable
$$COMMENT$$      display and solicit comments
$$CONTENT$$      output the HTML content of the section (TEMPLATE ONLY)
$$CONTENT$..$$   output the HTML of the specified section
$$CRUMBS$$       breadcrumb style navigation
$$DRAFT$$        if draft file in use (i.e. two underscores) output **DRAFT**
$$ELSE$$         see $$IF$$
$$ENDIF$$        see $$IF$$
$$FEEDBACK$$     feedback to email address
$$FORM$$         form elements (see FORM.C)
$$GOTO$$         redirect to this URL (entry in TOC)
$$GONETO$$       redirect to this URL (no entry in TOC)
$$IF$$           conditional to $$ELSE$$ and/or $$ENDIF$$ within a section
$$HIT$$          number of times section (page) hit
$$HTML$$         output raw HTML
$$INCLUDE$$      include (in an IFRAME) the supplied URL
$$LOGIN$$        login/logout dialog (see TemplateLogin() for three parameters)
$$MANAGE$$       provide the management interface
$$NEXT$$         link to next section
$$NOPRINT$$      do not print this section
$$NOTOC$$        no table-of-content entry
$$PRAGMA$$       set some quirky value
$$PREV$$         link to previous section
$$PRINT$$        print the current section
$$QANDA$$        question-and-answer challenge
$$RECAPTCHA$$    Google reCAPTCHA (incl Mailhide)
$$RESULT$$       output the results of a site search
$$RESULT$KEYWORD$$  output the keyword used to perform the search
$$ROTATE$$       rotate the content by the specific number of degrees
$$RSS$$          Really Simple Syndication feed
$$SEARCH$$       search dialog
$$SHOW$$         show the parameter
$$SITEMAP$$      output a site-map
$$SYNOPSIS$$     synopsis of the specified (sub)sections
$$TAG$$          tag a particular section
$$TICKET$$       a SPAMbot measure
$$TEASER$$       put a 'comment' into the site map
$$TIME$$         show the current date/time
$$TITLE$$        display the title of the current section (or other)
$$TOC$$          table of content (navigation menu)
$$TOC$CSS$$      pure CSS table of content (navigation menu)
$$TOC$DROP$$     table of content (navigation, drop-down <FORM> list)
$$TOP$$          link to top of page
$$USER$$         show the logged-in username


QUERY STRINGS
-------------
Some of these have the form NAME=<value>.
?BLOG            if single-button blog
?COMMENT         if POST method submit comment
?FEEDBACK        reader feedback
?KEYWORD         query (search) keyword
?LOGIN           authenticate
?LOGOUT          logout
?MANAGE          enter site management interface (if authorised)
?PRINT           render content and activate JavaScript print dialog
?_PRINT          render content as if about to print
?RESULT          search and display result
?RSS             supply the specified RSS feed
?VIEW            exit site management interface


COPYRIGHT
---------
Copyright (C) 2008-2011 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY
---------------
30-JUL-2011  MGD  bugfix; ContentLoad() .htc URL
23-JUL-2011  MGD  $$QANDA$$ and $$RECAPTCHA$$
02-MAY-2011  MGD  TemplateLoad() and TemplateContent() nested conditional
                    testing (per TES) 
                  TemplateDirective() consolidates template/content directive
                    processing into a single code-path
                  $$IF$query=<keyword>$$ tests query string keyword
15-APR-2010  MGD  TemplateTestIf() add regular expression
07-APR-2010  MGD  $$NOPRINT$$ and $$TAG$$
                  TemplateInclude() significant rework
26-JAN-2010  MGD  bugfix; TemplateTestIf()
25-NOV-2009  MGD  $$HTML$$ allows 'raw' HTML content
                  add "validate=" and "validator=" to $$RSS$$
22-NOV-2009  TES  bugfix; TemplateTestIf()
                  and add $$IF$$ equality operators
15-OCT-2009  MGD  TemplateUri() fix URI munging
19-SEP-2009  MGD  TemplateRss() and $$RSS$$
26-AUG-2009  MGD  TemplateInclude() VWCMS_SPAWN_<site-name> logical name
02-AUG-2009  MGD  $$TOC$ACROSS$$ superceded by $$TOC$CSS$$
10-APR-2009  MGD  TemplateSpawn() to allow HTML-escaping of output
                  lib$spawn() to disable authorised privileges
28-MAR-2009  MGD  $$TITLE$$ additional behaviour
                  $$TOC$ACROSS$$
24-MAR-2008  MGD  initial
*/

/*****************************************************************************/

#ifdef VWCMS_VMS_V62
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60200000
#undef __CRTL_VER
#define __CRTL_VER 60200000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <locale.h>
#include <stat.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <types.h>
#include <time.h>
#include <unistd.h>

/* VMS related header files */
#include "descrip.h"
#include "ssdef.h"
#include "stsdef.h"
#include "starlet.h"

/* application header file */
#include "ago.h"
#include "cgilib.h"
#include "config.h"
#include "comment.h"
#include "content.h"
#include "cookie.h"
#include "data.h"
#include "feedback.h"
#include "form.h"
#include "login.h"
#include "manage.h"
#include "other.h"
#include "qanda.h"
#include "recaptcha.h"
#include "regex.h"
#include "search.h"
#include "sendmail.h"
#include "synopsis.h"
#include "template.h"
#include "ticket.h"
#include "vwcms.h"

#define FI_LI "TEMPLATE", __LINE__

#define SUBS_BUFFER_CHUNK 4096

/* prototypes */

/* global storage */

BOOL  SmellsLikeMSIE567,
      TemplateIsDraft,
      TemplateLogout;

int  PragmaMsieIncludePercent = 100,
     TemplateIncludeCount,
     TemplateInsertContentCount,
     TemplateLineCount,
     TemplateTextLength;

char  *CgiHttpRefererPtr,
      *TemplateTextPtr;

char  TemplateFileName [256];

/* prototypes */

/* external storage */

extern BOOL  Debug,
             ContentIsDraft,
             ManageEditOn,
             ManageSiteOn,
             QueryComment,
             QueryFeedback,
             QueryLogin,
             QueryLogout,
             QueryKeyword,
             QueryPrint,
             QueryPrintPage,
             QueryPrintPagePlus,
             QueryPrintSite,
             QueryResult,
             QueryRss,
             QuerySubmit,
             RequestMethodPOST,
             RequestSubmitOK,
             WatchEnabled;

extern int  CommentTemplateCount,
            ManageAbsoluteX,
            ManageAbsoluteY;

extern ulong  ContentLastUpdated,
              CurrentSecondsTime,
              SectionLastUpdated;

extern ulong  CurrentVmsBinTime[];
extern ushort  CurrentVmsNumTime[];

extern char  *CgiHttpUserAgentPtr,
             *CgiRequestMethodPtr,
             *CgiRequestUriPtr,
             *ContentRssFeedNamePtr,
             *LastModifiedNow;

extern char  ContentFileName[],
             ContentQueryKeyword[],
             ContentSectionName[],
             ContentSiteName[],
             LinkSchemeHost[],
             LoginUserName[],
             SoftwareCopy[],
             SoftwareId[],
             Utility[];

extern CONTENT_DATA  *ContentSitePtr,
                     *ContentSectionPtr;

extern CONFIG_DATA  VWcmsConfig;

/*****************************************************************************/
/*
Load and parse the site template file.
*/

void TemplateLoad ()

{
   static DYSTRING  SubsBuffer = DYSTRING_EMPTY;

   BOOL  TemplateIf [IF_MAX+1];
   int  status,
        TemplateIfCount;
   char  chover;
   char  *aptr, *cptr, *sptr, *zptr,
         *EndHeadPtr = NULL;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateLoad()\n");

   /* obviously some initialisations */
   TemplateIsDraft = TemplateLogout = FALSE;
   CommentTemplateCount = TemplateIncludeCount = TemplateLineCount = 0;
   memset (TemplateIf, 0, sizeof(TemplateIf));
   TemplateIf[TemplateIfCount=0] = TRUE;

   zptr = (sptr = TemplateFileName) + sizeof(TemplateFileName)-1;
   if (QueryPrint)
   {
      /* if nothing like a directory specification add that of the site */
      if (!strchr (VWcmsConfig.SitePrint, ':'))
         for (cptr = VWcmsConfig.SiteDirectory;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
      for (cptr = VWcmsConfig.SitePrint;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      ManageEditOn = FALSE;
   }
   else
   {
      /* ditto */
      if (!strchr (VWcmsConfig.SiteTemplate, ':'))
         for (cptr = VWcmsConfig.SiteDirectory;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
      for (cptr = VWcmsConfig.SiteTemplate;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
   }
   if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
   *sptr = '\0';

   if (WatchEnabled) WatchThis (FI_LI, "TEMPLATE !AZ", TemplateFileName);

   status = DataReadFileIntoMemory (TemplateFileName,
                                    &TemplateTextPtr,
                                    &TemplateTextLength,
                                    TRUE);
   if (VMSnok (status))
   {
      CgiLibResponseError (FI_LI, status, TemplateFileName);
      return;
   }

   TicketTest (TRUE);
   RequestSubmitOK = FALSE;

   if (RequestMethodPOST)
   {
      /*********/
      /* POSTs */
      /*********/

      if (QueryComment)
      {
         if (GetCgiVarNull ("FORM_COMMENT_MODERATE"))
         {
            CommentModerate ();
            /* try to prevent reloading POSTs which might duplicate actions */
            for (cptr = sptr = CgiRequestUriPtr; *sptr && *sptr != '?'; sptr++);
            *sptr = '\0';
            CgiLibResponseRedirect ("%!%s\n%s", cptr, CookieHeader());
            return;
         }
         else
         if (GetCgiVarNull ("FORM_COMMENT_MAKE"))
         {
            if (CommentSubmit ()) return;
         }
         else
            ErrorExit (SS$_BUGCHECK, FI_LI);
      }
      else
      if (QueryFeedback)
      {
         if (GetCgiVarNull ("FORM_FEEDBACK_SUBMIT"))
         {
            if (FeedbackSubmit ()) return;
         }
         else
            ErrorExit (SS$_BUGCHECK, FI_LI);
      }
      else
      if (QuerySubmit)
      {
         if (FormSubmit (TRUE)) return;
      }
   }

   /******************/
   /* begin response */
   /******************/

   if (strstr (CgiHttpUserAgentPtr, "MSIE 7") ||
       strstr (CgiHttpUserAgentPtr, "MSIE 6") ||
       strstr (CgiHttpUserAgentPtr, "MSIE 5"))
   {
      SmellsLikeMSIE567 = TRUE;
      chover = '<';
   }
   else
   {
      SmellsLikeMSIE567 = FALSE;
      chover = 0;
   }

   CgiLibResponseHeader (200, "text/html",
"Last-Modified: %s\n\
Expires: Fri, 13 Jan 1978 14:00:00 GMT\n\
Cache-Control: no-cache, no-store, private, max-age=0, max-stale=0, \
must-revalidate, pre-check=0, post-check=0\n\
Pragma: no-cache\n%s",
                         LastModifiedNow,
                         CookieHeader());

   for (cptr = TemplateTextPtr; *cptr && isspace(*cptr); cptr++);

   /*
      The only way I seemed to be able to get TinyMCE to behave itself
      under MSIE 5, 6 and 7 and adopt the constraint of the parent container
      (textarea inside a table inside a div) was to switch the document (page)
      into "quirks mode" when editing!
   */
   if (*cptr == '<')
   {
      if (ManageEditOn &&
          SmellsLikeMSIE567 &&
          strsame (cptr, "<!DOCTYPE", 9))
      {
         /* MSIE - "quirks mode" here we come! */
         while (*cptr && *cptr != '>') cptr++;
         if (*cptr) cptr++;
         fputs ("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 \
Transitional//EN\">", stdout);
      }
   }

   sptr = cptr;
   while (*cptr)
   {
      while (*cptr && *cptr != chover &&
             *(USHORTPTR)cptr != '$$' &&
             *(USHORTPTR)cptr != '\\\n') cptr++;
      if (!*cptr) break;

      if (*cptr == '<')
      {
         /* only ever if 'chover' is '<' */
         if (strsame (cptr, "</HEAD", 6))
         {
            chover = 0;
            /* look ahead to see if pure-CSS TOC is in use */
            for (aptr = cptr+6; *aptr; aptr++)
            {
               if (*(USHORTPTR)aptr != '$$') continue;
               if (*(ULONGPTR)aptr != '$$to' &&
                   *(ULONGPTR)aptr != '$$TO') continue;
               if (strsame (aptr, "$$TOC$", 6)) break;
            }
            if (strsame (aptr, "$$TOC$CSS$", 10))
            {
               /* see CONTENT.C, ContentProcess() and ContentFileDot() */
               OutputFromTo (sptr, cptr);
               sptr = cptr;
               fprintf (stdout,
"<style type=\"text/css\" media=\"screen\">\
body{behavior:url(%s_csshover.htc);}\
</style>\n",
                        ContentEncodeName("/_",NULL));
               cptr += 6;
               continue;
            }
         }
         cptr++;
         continue;
      }

      if (TemplateIf[TemplateIfCount]) OutputFromTo (sptr, cptr);

      if (*(USHORTPTR)cptr == '\\\n')
      {
         /* suppressed newline */
         cptr += 2;
         sptr = cptr;
         continue;
      }

      if (*(ULONGPTR)cptr == '$$$$')
      {
         /* escaped double-dollar */
         if (TemplateIf[TemplateIfCount]) fputs ("$$", stdout);
         sptr = (cptr += 4);
         continue;
      }

      if (*(cptr+2) == '!')
      {
         /* commented-out section */
         cptr += 3;
         while (*cptr)
         {
            while (*cptr && *cptr != '!') cptr++;
            if (*cptr) cptr++;
            if (*(USHORTPTR)cptr != '$$') continue;
            cptr += 2;
            if (*cptr == '\n') cptr++;
            break;
         }
         sptr = cptr;
         continue;
      }

      /**************************/
      /* substitution directive */
      /**************************/

      /* using image-persistent storage */
      if (!DYSTRING_SIZE(&SubsBuffer)) DYSTRING_INIT(&SubsBuffer,0);

      cptr = TemplateParseDirect (&SubsBuffer, cptr);

      sptr = DYSTRING_PTR(&SubsBuffer);
      if (WatchEnabled) WatchThis (FI_LI, "SUBSTITUTE |!AZ", sptr);

      if (strsame (sptr, "IF", 2))
      {
         if (TemplateIfCount >= IF_MAX)
            TemplateReportError ("IF", NULL, IF_MAX_ERROR);
         else
         if (TemplateIf[TemplateIfCount])
            TemplateIf[++TemplateIfCount] = TemplateTestIf (sptr+2);
         else
            TemplateIf[++TemplateIfCount] = FALSE;
      }
      else
      if (strsame (sptr, "ENDIF", -1))
      {
         if (TemplateIfCount == 0)
            TemplateReportError ("ENDIF", NULL, IF_ENDIF_ERROR);
         else
            TemplateIf[TemplateIfCount--] = FALSE;
      }
      else
      if (strsame (sptr, "ELSE", -1))
      {
         if (TemplateIfCount == 0)
            TemplateReportError ("ELSE", NULL, IF_ELSE_ERROR);
         else
         if (TemplateIf[TemplateIfCount-1])
            TemplateIf[TemplateIfCount] = !TemplateIf[TemplateIfCount];
      }
      else
      if (TemplateIf[TemplateIfCount])
         TemplateDirective (sptr, FALSE);

      sptr = cptr;
   }

   if (TemplateIf[TemplateIfCount]) OutputFromTo (sptr, cptr);
}

/*****************************************************************************/
/*
If an error, warning or informational message has been generated then display
that via this substitution.  Once displayed the message is reset.  If the
'FaoString' parameter is NULL then place any advisory into a cookie for use by
the subsequent request.
*/

void TemplateAdvise
(
char *FaoString,
...
)
{
   static char  AdviseBuffer [1024];

   int  argcnt, status;
   ushort  ShortLength;
   ulong  *vecptr;
   ulong  FaoVector [64];
   char  *cptr, *sptr, *zptr,
         *AdvisePtr,
         *ErrorPtr,
         *InformPtr,
         *WarnPtr;
   char  ErrorString [256],
         InformString [256],
         Sandwich1 [256],
         Sandwich2 [256],
         WarnString [256];
   $DESCRIPTOR (BufferDsc, AdviseBuffer);
   $DESCRIPTOR (FaoStringDsc, FaoString);
   va_list  argptr;

   /*********/
   /* begin */
   /*********/

   va_count (argcnt);

   if (Debug)
      fprintf (stdout, "TemplateAdvise() %d |%s|\n", argcnt, FaoString);

   if (!FaoString)
   {
      /**********/ 
      /* cookie */
      /**********/ 

      /* place any advisory into a cookie */
      if (!AdviseBuffer[0]) return;
      cptr = Base64Encode (AdviseBuffer, -1);
      CookieSet (ADVISORY_COOKIE, cptr);
      AdviseBuffer[0] = '\0';
      return;
   }

   if (!FaoString[0] || FaoString[0] == '$')
   {
      /**************/
      /* $$advise$$ */
      /**************/

      /* the '$$$$' is the format string supplied by ManageInterface() */
      if (ManageEditOn && *(ULONGPTR)FaoString != '$$$$') return;

      if (WatchEnabled) WatchThis (FI_LI, "$$ADVISE$$ |!AZ", AdviseBuffer);

      if (!AdviseBuffer[0]) return;

      if (*(cptr = FaoString)) cptr++;

      zptr = (sptr = InformString) + sizeof(InformString)-1;
      while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';
      if (!*(InformPtr = InformString))
         InformPtr = CONFIG_LANG_PARSE(LangAdvise,1);

      if (*cptr == '$') cptr++;
      zptr = (sptr = WarnString) + sizeof(WarnString)-1;
      while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';
      if (!*(WarnPtr = WarnString))
         WarnPtr = CONFIG_LANG_PARSE(LangAdvise,2);

      if (*cptr == '$') cptr++;
      zptr = (sptr = ErrorString) + sizeof(ErrorString)-1;
      while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';
      if (!*(ErrorPtr = ErrorString))
         ErrorPtr = CONFIG_LANG_PARSE(LangAdvise,3);

      /* sandwiches 1 and 2 are used internally by ManageInterface() */
      if (*cptr == '$') cptr++;
      zptr = (sptr = Sandwich1) + sizeof(Sandwich1)-1;
      while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';
      if (*cptr == '$') cptr++;
      zptr = (sptr = Sandwich2) + sizeof(Sandwich2)-1;
      while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';

      switch (AdviseBuffer[0])
      {
         case 'M' : cptr = "vwcms_manage"; AdvisePtr = InformPtr; break; 
         case 'I' : cptr = "vwcms_inform"; AdvisePtr = InformPtr; break; 
         case 'W' : cptr = "vwcms_warn"; AdvisePtr = WarnPtr; break;
         case 'E' : default: cptr = "vwcms_error"; AdvisePtr = ErrorPtr;
      }
      sptr = HtmlEscape (AdviseBuffer+1);
      fprintf (stdout, "%s<span class=\"%s\">%s:&nbsp;%s</span>%s",
               Sandwich1, cptr, AdvisePtr, sptr, Sandwich2);
      AdviseBuffer[0] = '\0';
      return;
   }

   /*********/
   /* store */
   /*********/

   /* don't overwrite existing higher value error message */
   if (AdviseBuffer[0] == 'E' ||
       AdviseBuffer[0] == FaoString[0] ||
       (AdviseBuffer[0] == 'W' && FaoString[0] != 'E')) return;

   if (argcnt > 64) ErrorExit (SS$_BUGCHECK, FI_LI);

   vecptr = FaoVector;
   va_start (argptr, FaoString);
   for (argcnt -= 1; argcnt; argcnt--)
      *vecptr++ = (ulong)va_arg (argptr, ulong);
   va_end (argptr);

   FaoStringDsc.dsc$w_length = strlen(FaoString)+1;
   status = sys$faol (&FaoStringDsc, &ShortLength, &BufferDsc,
                      (ulong*)&FaoVector);
   if (Debug) fprintf (stdout, "sys$fao() %%X%08.08X\n", status);

   if (!(status & 1)) sprintf (AdviseBuffer, "E$FAO() %%X%08.08X\n", status); 

   if (WatchEnabled) WatchThis (FI_LI, "ADVISE |!AZ", AdviseBuffer);
}

/*****************************************************************************/
/*
The content last-modified date/time in days/hours/minutes from current time.
Also see AgoLangInit().

If the first parameter is '#' the provide the section last-modified.
If empty or '*' then provide the document last-modified.
*/

void TemplateAgo (char *SubsArg)

{
   char  *cptr;
   ulong  LastUpdated;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateAgo()\n");

   if (*(cptr = SubsArg) == '$') cptr++;

   if (*cptr && *cptr == '#') {
      if (!(LastUpdated = SectionLastUpdated))
         LastUpdated = ContentLastUpdated; 
   }
   else
      LastUpdated = ContentLastUpdated; 

   fputs (AgoWas(TimeString("VMS20",LastUpdated)), stdout);
}

/*****************************************************************************/
/*
Only available with JavaScript.
Of course the browsers diasagree on where to start counting!!
*/

void TemplateBack (char* SubsArg)

{
   char  *cptr, *sptr, *zptr,
         *BackPtr;
   char  BackString [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateBack() |%s|\n", SubsArg);

   if (*SubsArg == '$') SubsArg++; else SubsArg = "";
   cptr = SubsArg;
   zptr = (sptr = BackString) + sizeof(BackString)-1;
   while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';

   if (!*(BackPtr = BackString))
      BackPtr = CONFIG_LANG_PARSE(LangNavigate,3);

   if (ManageEditOn)
   {
      if (cptr = IsAnImage (BackPtr))
         fprintf (stdout, "<img class=\"vwcms_nav_img_off\" \
alt=\"[back]\" %s />", cptr);
      else
         fprintf (stdout, "<span class=\"vwcms_nav\">%s</span>", BackPtr);
   }
   else
   {
      fprintf (stdout,
"<script type=\"text/javascript\">\
if(/Gecko[\\/\\s](\\d+)/.test(navigator.userAgent)){hempty=1}else{hempty=0};");

      if (cptr = IsAnImage (BackPtr))
         fprintf (stdout,
"if(history.length>hempty){document.write(\'\
<a%s href=\"javascript:history.go(-1)\">\
<img class=\"vwcms_nav_img\" alt=\"[back]\" %s /></a>\')}\
else{document.write(\'<img class=\"vwcms_nav_img_off\" \
alt=\"[back]\" %s />\')}\
</script><noscript><img class=\"vwcms_nav_img_off\" \
alt=\"[back]\" %s /></noscript>",
                  ManageLeaveNow(),
                  cptr, cptr, cptr);
      else
         fprintf (stdout,
"if(history.length>hempty){document.write(\'\
<span class=\"vwcms_nav\">\
<a%s href=\"javascript:history.go(-1)\">%s</a></span>\')}\
else{document.write(\'<span class=\"vwcms_nav\">%s</span>\')}\
</script><noscript><span class=\"vwcms_nav\">%s</span></noscript>",
                  ManageLeaveNow(),
                  BackPtr, BackPtr, BackPtr);
   }
}

/*****************************************************************************/
/*
Facilitate weblogging.
See SynopsisContent() for parameter details.
The new blog entry emulates a [H1]..[Hn] button on the content management page
resulting in a new section being inserted and the content editor to be opened.
*/

void TemplateBlog (char *SubsArg)

{
   int  retval,
        SectionLevel = 2;
   ulong  TimeStampSeconds;
   char  *cptr, *sptr,
         *BlogNewPtr,
         *BlogTitlePtr,
         *FormatPtr = NULL,
         *LevelArgPtr = "level=0";
   char  BlogTitle [256] = "",
         BlogNew [256] = "",
         FormatString [256],
         TimeStamp [256];
   struct tm  time_struct;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateBlog() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;

   if (!*cptr)
   {
      TemplateParamError ("BLOG", NULL, NULL);
      return;
   }

   if (strsame (cptr, "new", 3))
   {
      /******************/
      /* new blog entry */
      /******************/

      /* only shown if logged-in and capable of editing */
      if (!ManageSiteOn) return;

      cptr = TemplateParseParam (cptr+3, BlogNew, sizeof(BlogNew));
      while (*cptr)
      {
         if (strsame (cptr, "title=", 6))
            cptr = TemplateParseParam (cptr+6, BlogTitle, sizeof(BlogTitle));
         else
         {
            TemplateParamError ("BLOG", SubsArg, cptr);
            return;
         }
      }   

      SectionLevel = ContentSectionPtr->HeadNumber + 1;
      if (SectionLevel > VWcmsConfig.SiteDepth)
      {
         TemplateReportError ("BLOG", NULL, "too deep");
         return;
      }

      if (!*(BlogNewPtr = BlogNew))
         BlogNewPtr = CONFIG_LANG_PARSE(LangBlog,2);

      if (!*(BlogTitlePtr = BlogTitle))
         BlogTitlePtr = CONFIG_LANG_PARSE(LangBlog,3);

      for (cptr = ContentFileName; *cptr; *cptr++);
      while (cptr > ContentFileName && *cptr != ']') cptr--;
      if (*cptr == ']') cptr++;

      if (ContentSectionPtr)
         sptr = ContentSectionPtr->SectionName;
      else
         sptr = "$$EOF$$";

      fprintf (stdout,
"<div class=\"vwcms_blog_form\">\n\
<form class=\"vwcms_blog\" id=\"blog_form\" name=\"blog_form\" \
method=\"POST\" action=\"%s?blog\">\n\
<input type=\"hidden\" name=\"contentfile\" value=\"%s\" />\n\
<input type=\"hidden\" name=\"currentsect\" value=\"%s\" />\n\
<nobr>\
<input class=\"vwcms_btn\" style=\"margin-top:1em;\" \
id=\"blog_submit\" name=\"h%d\" ",
               ContentEncodeName("/",ContentSectionName),
               cptr,
               sptr,
               SectionLevel);

      if (cptr = IsAnImage(BlogNewPtr))
         fprintf (stdout, "type=\"image\" %s%s />",
                  cptr, RequestSubmitOK ? " disabled=\"disabled\"" : "");
      else
         fprintf (stdout, "type=\"submit\" value=\"%s\"%s />",
                  BlogNewPtr, RequestSubmitOK ? " disabled=\"disabled\"" : "");

      fprintf (stdout,
"&nbsp;<input style=\"margin-top:1em;\" name=\"sectitle\" \
onfocus=\"this.select()\" \
onkeyup=\"document.getElementById(\'blog_submit\').disabled=false;\" \
type=\"text\" size=\"%d\" value=\"%s\" />\
</nobr>\n\
</form>\n\
</div>\n",
               48,
               BlogTitlePtr);

      return;
   }

   /* does it look like it could be a time string rather than keyword? */
   for (sptr = cptr;
        *sptr && *sptr != '=' && *sptr != '$' && *sptr != '.';
        sptr++);
   if (*sptr != '=' && *sptr != '.')
   {
      /*******************/
      /* entry date/time */
      /*******************/

      cptr = TemplateParseParam (cptr, TimeStamp, sizeof(TimeStamp));
      if (*cptr == '%')
         cptr = TemplateParseParam (cptr,
                                    FormatPtr = FormatString,
                                    sizeof(FormatString));

      sptr = GetTimeString (TimeStamp, &time_struct);
      TimeStampSeconds = mktime (&time_struct);
      if ((int)TimeStampSeconds == -1) sptr = NULL;

      if (sptr)
         fprintf (stdout, "<span class=\"vwcms_timestamp\">%s</span>",
                  TimeString(FormatPtr,TimeStampSeconds));
      else
         TemplateReportError ("BLOG", TimeStamp, "strptime()?");

      return;
   }

   /*******************/
   /* list of entries */
   /*******************/

   SynopsisContent (NULL, SubsArg);
}

/*****************************************************************************/
/*
Substitute the value of the CGI variable (always HTML-escaped).
*/

void TemplateCgi (char *SubsArg)

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateCgi() |%s|\n", SubsArg);

   if (!*SubsArg)
   {
      TemplateParamError ("CGI", NULL, NULL);
      return;
   }
   if (!(cptr = GetCgiVarNull (SubsArg+1))) return;
   fputs (HtmlEscape(cptr), stdout);
}

/*****************************************************************************/
/*
Output the section content.  This can be a 'special' page such as LOGIN. 
It can also be the VWCMS management interface around that page.  Generally
though it's just the HTML content of the requested site section.  This needs to
be parsed for the limited number of 'special case' substitution sequences, such
as "$$SITEMAP$$", that are supported in the content.
*/

void TemplateContent (CONTENT_DATA *SectionPtr)

{
   static DYSTRING  SubsBuffer = DYSTRING_EMPTY;

   BOOL  ContentIf [IF_MAX+1];
   int  ContentIfCount;
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateContent()\n");

   memset (ContentIf, 0, sizeof(ContentIf));
   ContentIf[ContentIfCount=0] = TRUE;

   if (QueryKeyword)
   {
      if (cptr = GetCgiVarNull ("FORM_KEYWORD"))
      {
         /* ensure it's not empty */
         while (*cptr && isspace(*cptr)) cptr++;
         if (!*cptr) QueryKeyword = FALSE;
      }
   }

   cptr = NULL;

   if (SectionPtr)
   {
      /* section supplied as argument */
      cptr = SectionPtr->TextPtr;
   }
   else
   if (QueryLogin)
   {
      if ((SectionPtr = ContentFindSection ("$$LOGIN$$")) &&
          SectionPtr->TextLength)
         cptr = SectionPtr->TextPtr;
      else
         fputs ("$$LOGIN$$ \?\?\?", stdout);
   }
   else
   if (QueryKeyword)
   {
      if ((SectionPtr = ContentFindSection ("$$RESULT$$")) &&
          SectionPtr->TextLength)
         cptr = SectionPtr->TextPtr;
      else
         fputs ("$$RESULT$$ \?\?\?", stdout);
   }
   else
   if (ManageEditOn)
      ManageInterface ();
   else
   if (ContentSectionPtr)
   {
      /* section specified as global data */
      cptr = ContentSectionPtr->TextPtr;
   }
   else
   if (VWcmsConfig.Unknown302)
   {
      /* section not found and redirection configured */
      Unknown302();
      return;
   }
   else
   if ((SectionPtr = ContentFindSection ("$$UNKNOWN$$")) &&
       SectionPtr->TextLength)
      cptr = SectionPtr->TextPtr;
   else
      fputs ("$$UNKNOWN$$ \?\?\?", stdout);

   if (cptr)
   {
      /**********************/
      /* output the content */
      /**********************/

      if (!*cptr)
      {
         fprintf (stdout, "<div class=\"vwcms_no_content\">%s</div>",
                  CONFIG_LANG(LangEmpty));
         return;
      }

      sptr = cptr;
      while (*cptr)
      {
         while (*cptr && *cptr != '\n' && *(USHORTPTR)cptr != '$$') cptr++;
         if (!*cptr) break;

         if (*cptr == '\n')
         {
            cptr++;
            if (!QueryResult) continue;
            OutputFromTo (sptr, cptr);
            fprintf (stdout, "<a name=\"l%d\"></a>", ++TemplateLineCount);
            sptr = cptr;
            continue;
         }

         if (ContentIf[ContentIfCount]) OutputFromTo (sptr, cptr);

         if (*(ULONGPTR)cptr == '$$$$')
         {
            /* escaped double-dollar */
            if (ContentIf[ContentIfCount]) fputs ("$$", stdout);
            sptr = (cptr += 4);
            continue;
         }

         if (*(cptr+2) == '!')
         {
            /* commented-out section */
            cptr += 3;
            while (*cptr)
            {
               while (*cptr && *cptr != '!') cptr++;
               if (*cptr) cptr++;
               if (*(USHORTPTR)cptr != '$$') continue;
               cptr += 2;
               if (*cptr == '\n') cptr++;
               break;
            }
            sptr = cptr;
            continue;
         }

         /**************************/
         /* substitution directive */
         /**************************/

         /* using image-persistent storage */
         if (!DYSTRING_SIZE(&SubsBuffer)) DYSTRING_INIT(&SubsBuffer,0);

         cptr = TemplateParseDirect (&SubsBuffer, cptr);

         sptr = DYSTRING_PTR(&SubsBuffer);
         if (WatchEnabled) WatchThis (FI_LI, "SUBSTITUTE |!AZ", sptr);

         if (strsame (sptr, "IF", 2))
         {
            if (ContentIfCount >= IF_MAX)
               TemplateReportError ("IF", NULL, IF_MAX_ERROR);
            else
            if (ContentIf[ContentIfCount])
               ContentIf[++ContentIfCount] = TemplateTestIf (sptr+2);
            else
               ContentIf[++ContentIfCount] = FALSE;
         }
         else
         if (strsame (sptr, "ENDIF", -1))
         {
            if (ContentIfCount == 0)
               TemplateReportError ("ENDIF", NULL, IF_ENDIF_ERROR);
            else
               ContentIf[ContentIfCount--] = FALSE;
         }
         else
         if (strsame (sptr, "ELSE", -1))
         {
            if (ContentIfCount == 0)
               TemplateReportError ("ELSE", NULL, IF_ELSE_ERROR);
            else
            if (ContentIf[ContentIfCount-1])
               ContentIf[ContentIfCount] = !ContentIf[ContentIfCount];
         }
         else
         if (ContentIf[ContentIfCount])
            TemplateDirective (sptr, TRUE);

         sptr = cptr;
      }

      if (ContentIf[ContentIfCount]) OutputFromTo (sptr, cptr);

      /* finalise anything that hasn't been */
      FormTemplate (NULL);
   }
}

/*****************************************************************************/
/*
Process (many) $$<directive>$$.
*/

char  *TemplateContentDirectives [] =
{ "ADVISE", "AGO", "BACK", "BLOG", "CGI", "COMMENT", "CONTENT$", "CRUMBS",
  "DRAFT", "ENDIF", "ELSE", "FEEDBACK", "GOTO", "GONETO", "HIDE", "HIT",
  "HTML", "IF", "INCLUDE", "INSERT", "LASTMOD", "NEXT", "NOPRINT", "NOTOC",
  "PRAGMA", "PREV", "PRINT", "QANDA", "RECAPTCHA", "RESULT", "RSS", "SECTION",
  "SHOW", "SITEMAP", "STATS", "SYNOPSIS", "TEASER", "TAG", "TICKET", "TIME",
  "USER",
  NULL };

void TemplateDirective
(
char *sptr,
BOOL InContent
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateDirective() %d\n", InContent);

   if (*sptr == '_')
      TemplateMacro (sptr);
   else
   if (strsame (sptr, "ADVISE", 6) ||
       strsame (sptr, "ADVICE", 6))
      TemplateAdvise (sptr+6);
   else
   if (strsame (sptr, "AGO", 3))
      TemplateAgo (sptr+3);
   else
   if (strsame (sptr, "BACK", 4))
      TemplateBack (sptr+4);
   else
   if (strsame (sptr, "BLOG", 4))
      TemplateBlog (sptr+4);
   else
   if (strsame (sptr, "CGI", 3))
      TemplateCgi (sptr+3);
   else
   if (strsame (sptr, "CLIP", 4))
   {
      /* will happen when a section is removed into a clip */
      fprintf (stdout, "$$%s$$", sptr);
   }
   else
   if (strsame (sptr, "COMMENT", 7))
      CommentTemplate (sptr+7);
   else
   if (strsame (sptr, "CONTENT$", 8))
   {
      /* must preceded the "CONTENT" test */
      TemplateInsertContent (sptr+7);
   }
   else
   if (strsame (sptr, "CONTENT", -1))
   {
      /* must follow the "CONTENT$" test */
      if (!InContent)
      {
         /* only allowed in template */
         if (QueryPrintPage)
            TemplateRange (0);
         else
         if (QueryPrintPagePlus)
            TemplateRange (1);
         else
         if (QueryPrintSite)
            TemplateRange (9);
         else
         if (GetCgiVarNull ("FORM_EXPORT"))
            TemplateRange (9);
         else
            TemplateContent (NULL);
      }
      else
         TemplateReportError (sptr, NULL, CONFIG_LANG_PARSE(LangSubsError,1));
   }
   else
   if (strsame (sptr, "CRUMBS", 6))
      ContentCrumbs (sptr+6);
   else
   if (strsame (sptr, "DRAFT", 5))
      TemplateDraft (sptr+5);
   else
   if (strsame (sptr, "FEEDBACK", 8))
      FeedbackTemplate (sptr+8);
   else
   if (strsame (sptr, "GOTO", 4))
   {
      if (InContent)
      {
         /* should never be encountered during content processing */
         TemplateReportError ("GOTO", NULL, "1.0.0.0.0");
      }
   }
   else
   if (strsame (sptr, "GONETO", 6))
   {
      if (InContent)
      {
         /* should never be encountered during content processing */
         TemplateReportError ("GONETO", NULL, "1.0.0.0.0");
      }
   }
   else
   if (strsame (sptr, "FORM", 4))
      FormTemplate (sptr+4);
   else
   if (strsame (sptr, "HIDE", -1))
      /* detected and processed by ContentParseLine1() */
      sptr = NULL;
   else
   if (strsame (sptr, "HIT", 3))
      TemplateHit (sptr+3);
   else
   if (strsame (sptr, "HTML", 4))
      TemplateHtml (sptr+4);
   else
   if (strsame (sptr, "INCLUDE", 7))
      TemplateInclude (FALSE, sptr+7);
   else
   if (strsame (sptr, "INSERT", 6))
      TemplateInclude (TRUE, sptr+6);
   else
   if (strsame (sptr, "LASTMOD", 7))
      TemplateLastModified (sptr+7);
   else
   if (strsame (sptr, "LOGIN", 5))
      TemplateLoginLogout (sptr+5);
   else
   if (strsame (sptr, "MANAGE", 6))
      TemplateManage (sptr+6);
   else
   if (strsame (sptr, "NEXT", 4))
      TemplateNext (sptr+4);
   else
   if (strsame (sptr, "NOPRINT", 7))
      /* detected and processed by ContentParseLine1() */
      sptr = NULL;
   else
   if (strsame (sptr, "NOTOC", -1))
      /* detected and processed by ContentParseLine1() */
      sptr = NULL;
   else
   if (strsame (sptr, "PRAGMA", 6))
      TemplatePragma (sptr+6);
   else
   if (strsame (sptr, "PREV", 4))
      TemplatePrev (sptr+4);
   else
   if (strsame (sptr, "PRINT", 5))
      TemplatePrint (sptr+5);
   else
   if (strsame (sptr, "QANDA", 5))
      QandaDirective (sptr+5);
   else
   if (strsame (sptr, "RECAPTCHA", 9))
      ReCaptchaDirective (sptr+9);
   else
   if (strsame (sptr, "RESULT", 6))
      SearchResult (sptr+6);
   else
   if (strsame (sptr, "ROTATE", 6))
      TemplateRotate (sptr+6);
   else
   if (strsame (sptr, "RSS", 3))
      TemplateRSS (sptr+3);
   else
   if (strsame (sptr, "SEARCH", 6))
      TemplateSearch (sptr+6);
   else
   if (strsame (sptr, "SHOW", 4))
      TemplateShow (sptr+4);
   else
   if (strsame (sptr, "SITEMAP", -1))
      ContentSiteMap ();
   else
   if (strsame (sptr, "STATS", 5))
      TemplateStats (sptr+5);
   else
   if (strsame (sptr, "SYNOPSIS", 8))
      SynopsisContent (sptr+8, NULL);
   else
   if (strsame (sptr, "TAG", 3))
      ContentTag (sptr+3);
   else
   if (strsame (sptr, "TEASER", 6))
      /* detected and processed by ContentSiteMap() */
      sptr = NULL;
   else
   if (strsame (sptr, "TICKET", 6))
      TicketIssue (sptr+6);
   else
   if (strsame (sptr, "TIME", 4))
      TemplateTime (sptr+4, 0);
   else
   if (strsame (sptr, "TITLE", 5))
      TemplateTitle (sptr+5);
   else
   if (strsame (sptr, "TOC", 3))
   {
      /* only allowed in template */
      if (!InContent)
      {
         /* must go before "toc" for the obvious reason */
         if (strsame (sptr, "TOC$CSS", 7))
            ContentTocCss (sptr+7);
         else
         /* ditto */
         if (strsame (sptr, "TOC$DROP", 8))
            ContentTocDrop (sptr+8);
         else
         if (strsame (sptr, "TOC$PRINT", 9))
            ContentTocPrint (sptr+9);
         else
            ContentTocInline (sptr+3);
      }
      else
         TemplateReportError (sptr, NULL, CONFIG_LANG_PARSE(LangSubsError,1));
   }
   else
   if (strsame (sptr, "TOP", 3))
      TemplateTop (sptr+3);
   else
   if (strsame (sptr, "USER", 4))
      TemplateUser (sptr+4);
   else
      TemplateReportError (sptr, NULL, CONFIG_LANG_PARSE(LangSubsError,1));
}

/*****************************************************************************/
/*
If it is a draft begin viewed (the user is logged-in and the content file
begins with two underscores) then display "***DRAFT***" or whatever message is
supplied.
*/

void TemplateDraft (char *SubsArg)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateDraft() |%s|\n", SubsArg);

   if (!ContentIsDraft) return;

   if (*SubsArg == '$') SubsArg++; else SubsArg = "";
   if (!*SubsArg) SubsArg = CONFIG_LANG_PARSE(LangManage2,4);
   fprintf (stdout, "%s ", SubsArg);
}

/*****************************************************************************/
/*
*/

void TemplateHit (char *SubsArg)

{
   int  status,
        HitCount;
   char  ch;
   char  *cptr, *sptr,
         *CounterNamePtr,
         *FileNamePtr,
         *NamePtr;
   char  CounterName [256] = "",
         SinceWhen [256] = "",
         StringBuffer [256],
         TimeBuffer [256];
   stat_t  StatBuffer;
   struct tm  *tm_ptr;
   FILE  *FilePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateHit() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;

   while (*cptr)
   {
      if (*cptr == '%')
         cptr = TemplateParseParam (cptr, SinceWhen, sizeof(SinceWhen));
      else
         cptr = TemplateParseParam (cptr, CounterName, sizeof(CounterName));
   }

   if (CounterName[0])
      CounterNamePtr = CounterName;
   else
   if (ContentSectionPtr)
      CounterNamePtr = ContentSectionPtr->SectionName;
   else
      CounterNamePtr = ContentSectionName;

   FileNamePtr = MakeFileName (CounterNamePtr, ".HIT");

   status = WaitForLock (FileNamePtr, TRUE);
   if (VMSnok (status))
   {
      TemplateAdvise ("EHIT !AZ (%X!8XL)", SysGetMsg(status), status);
      return;
   }
                                          
   if (SinceWhen[0])
   {
      if (stat (FileNamePtr, &StatBuffer) < 0)
      {
         status = vaxc$errno;
         TemplateAdvise ("EHIT !AZ (%X!8XL)", SysGetMsg(status), status);
         return;
      }
      if (*(USHORTPTR)SinceWhen == '%\0') SinceWhen[0] = '\0';
      TemplateTime (SinceWhen, StatBuffer.st_ctime);
      WaitForUnLock ();
      return;
   }

   FilePtr = fopen (FileNamePtr, "w+", "ctx=stm", "rfm=stmlf",
                    "fop=cif", "shr=nil");
   if (!FilePtr)
   {
      status = vaxc$errno;
      TemplateAdvise ("EHIT !AZ (%X!8XL)", SysGetMsg(status), status);
      WaitForUnLock ();
      return;
   }

   rewind (FilePtr);
   StringBuffer[0] = '\0';
   fgets (StringBuffer, sizeof(StringBuffer)-1, FilePtr);
   HitCount = atoi(StringBuffer);
   rewind (FilePtr);
   fprintf (FilePtr, "%d\n", ++HitCount);

   fclose (FilePtr);

   WaitForUnLock ();

   fprintf (stdout, "%d", HitCount);
}

/*****************************************************************************/
/*
Although the HTML editor will HTML-entify the text as it is edited this just
reverses the process.  Careful!  The HTML editor will usually do two things;
1) insert its own markup for newlines, etc., 2) HTML-entify HTML-forbidden
characters.  This function removes all HTML markup from the supplied argument,
then HTML-deentifies it and outputs the content into the text.  This allows the
user to explicitly insert their own HTML using the HTML editor.  Of course the
entire parameter must be explicitly marked-up like any other HTML!
*/

void TemplateHtml (char* SubsArg)

{
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateHtml() |%s|\n", SubsArg);

   if (*SubsArg == '$') SubsArg++;

   for (cptr = SubsArg; *cptr && *cptr != '<'; cptr++);
   sptr = cptr;
   while (*cptr)
   {
      /* process non-tag text */
      while (*cptr && *cptr != '<') *sptr++ = *cptr++;
      while (*cptr)
      {
         /* process tag text */
         while (*cptr && *cptr != '>' && *cptr != '\"' && *cptr != '\'') cptr++;
         if (*cptr == '\"')
         {
            /* quoted tag entity */
            while (*cptr && *cptr != '\"') cptr++;
            if (*cptr) cptr++;
         }
         else
         if (*cptr == '\'')
         {
            /* quoted tag entity */
            while (*cptr && *cptr != '\'') cptr++;
            if (*cptr) cptr++;
         }
         if (*cptr == '>')
         {
            /* looks like the end of the tag */
            cptr++;
            break;
         }
      }
   }
   *sptr = '\0';

   if (Debug) fprintf (stdout, "|%s|\n", SubsArg);

   ContentDeEntify (SubsArg, FALSE);
   fputs (SubsArg, stdout);
}

/*****************************************************************************/
/*
Return true or false based on testing the parameter condition(s).
*/

BOOL TemplateTestIf (char *SubsArg)

{
   static int  RegexCflags = REG_EXTENDED | REG_ICASE | REG_NOSUB,
               RegexEflags = 0;

   BOOL  IsCgi,
         IsLnm,
         TestResult = FALSE,
         ThisResult,
         WithNot,
         WithAnd;
   int  retval;
   char  *aptr, *cptr, *sptr, *optr, *tptr, *zptr,
         *ParamPtr;
   char  CgiLnmName [256],
         KeyWord [256],
         RegExErrMsg [256],
         TestCond [255+3+255+1];  /* CgiLnmName + operator + KeyWord */
   regex_t  regexp;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateTestIf() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;

   while (*cptr)
   {
      ParamPtr = cptr;
      IsCgi = IsLnm = ThisResult = WithNot = WithAnd = FALSE;
      if (*(USHORTPTR)cptr == '&&')
      {
         WithAnd = TRUE;
         cptr += 2;
      }
      else
      if (*(USHORTPTR)cptr == '||')
      {
         WithAnd = FALSE;
         cptr += 2;
      }
      else
      if (*cptr == '!')
      {
         WithNot = TRUE;
         cptr++;
      }

      if (*(USHORTPTR)cptr == '1\0' ||
          *(USHORTPTR)cptr == '1$')
         ThisResult = TRUE;
      else
      if (*(USHORTPTR)cptr == '0\0' ||
          *(USHORTPTR)cptr == '0$')
         ThisResult = FALSE;
      else
      if ((IsCgi = strsame (cptr, "cgi=", 4)) ||
          (IsLnm = strsame (cptr, "lnm=", 4)))
      {
         /* To match the comparison operators it is necessary to de-entify the
          * substitution string, which means locating either the end of the
          * string or the next '$' and and calling CgiLibHtmlDeEntify().
          */
         zptr = (sptr = TestCond) + sizeof(TestCond)-1;
         for (tptr = cptr + 4;
              *tptr && *tptr != '$' && sptr < zptr;
              *sptr++ = *tptr++);
         *sptr = '\0';

         CgiLibHtmlDeEntify(TestCond);

         zptr = (sptr = CgiLnmName) + sizeof(CgiLnmName)-1;
         tptr = TestCond;
         while (*tptr && sptr < zptr)
         {
            if (*tptr == '!' || *tptr == '<' ||
                *tptr == '=' || *tptr == '>' || *tptr == '^')
               break;
            *sptr++ = *tptr++;
         }
         *sptr = '\0';

         if (IsCgi)
            aptr = GetCgiVarNull (CgiLnmName);
         else
            aptr = TrnLnm (CgiLnmName, NULL, 0);

         if (aptr && !*tptr)
            ThisResult = TRUE;
         else
         if (aptr)
         {
            optr = tptr;
            if (*tptr == '!' || *tptr == '<' ||
                *tptr == '=' || *tptr == '>' ||
                *tptr == '^') tptr++;
            if (*tptr == '=' || *tptr == '>') tptr++;
            for (zptr = (sptr = KeyWord) + sizeof(KeyWord)-1;
                 *tptr && sptr < zptr;
                 *sptr++ = *tptr++);
            *sptr = '\0';
            /* push to all lower case so we can use V6.0 strcmp() */
            for (tptr = aptr; *tptr; tptr++) *tptr = tolower(*tptr);
            for (tptr = KeyWord; *tptr; tptr++) *tptr = tolower(*tptr);
            ThisResult = FALSE;
            if (*(USHORTPTR)optr == '==')
               ThisResult = strcmp (aptr, KeyWord) == 0;
            else
            if (*(USHORTPTR)optr == '<=')
               ThisResult = strcmp (aptr, KeyWord) <= 0;
            else
            if (*(USHORTPTR)optr == '>=')
               ThisResult = strcmp (aptr, KeyWord) >= 0;
            else
            if (*(USHORTPTR)optr == '!=')
               ThisResult = strcmp (aptr, KeyWord) != 0;
            else
            if (*(USHORTPTR)optr == '<>')
               ThisResult = strcmp (aptr, KeyWord) != 0;
            else
            if (*optr == '<')
               ThisResult = strcmp (aptr, KeyWord) < 0;
            else
            if (*optr == '=')
               ThisResult = (BOOL) SearchKeyWord (aptr, KeyWord);
            else
            if (*optr == '>')
               ThisResult = strcmp (aptr, KeyWord) >= 0;
            else
            if (*optr == '^')
            {
               /* regular expression */
               retval = regcomp (&regexp, KeyWord, RegexCflags);
               if (retval)
               {
                  regerror (retval, &regexp, RegExErrMsg, sizeof(RegExErrMsg));
                  regfree (&regexp);
                  if (WatchEnabled) WatchThis (FI_LI, "REGEX !AZ", RegExErrMsg);
               }
               else
               {
                  retval = regexec (&regexp, aptr, 0, 0, RegexEflags);
                  regfree (&regexp);
               }
               ThisResult = (retval == 0);
            }
         }
         else
            ThisResult = FALSE;
      }
      else
      if (strsame (cptr, "login=", 6))
      {
         zptr = (sptr = KeyWord) + sizeof(KeyWord)-1;
         for (cptr += 6;
              *cptr && *cptr != '$' && sptr < zptr;
              *sptr++ = *cptr++);
         *sptr = '\0';
         CgiLibHtmlDeEntify(KeyWord);
         ThisResult = strsame (LoginUserName, KeyWord, -1);
      }
      else
      if (strsame (cptr, "login", -1) ||
          strsame (cptr, "login$", -1))
         ThisResult = LoginUserName[0];
      else
      if (strsame (cptr, "manage=edit", 11))
         ThisResult = ManageEditOn;
      else
      if (strsame (cptr, "manage", -1) ||
          strsame (cptr, "manage$", -1))
         ThisResult = ManageSiteOn;
      else
      if (strsame (cptr, "print=page+", 11))         
         ThisResult = QueryPrintPagePlus;
      else
      if (strsame (cptr, "print=page", 10))         
         ThisResult = QueryPrintPage;
      else
      if (strsame (cptr, "print=site", 10) ||
          strsame (cptr, "print=all", 9))         
         ThisResult = QueryPrintSite;
      else
      if (strsame (cptr, "print", -1) ||
          strsame (cptr, "print$", -1))
         ThisResult = QueryPrint;
      else
      if (strsame (cptr, "query=", 6))
      {
         zptr = (sptr = KeyWord) + sizeof(KeyWord)-1;
         for (cptr += 6;
              *cptr && *cptr != '$' && sptr < zptr;
              *sptr++ = *cptr++);
         *sptr = '\0';
         CgiLibHtmlDeEntify(KeyWord);
         ThisResult = strsame (ContentQueryKeyword, KeyWord, -1);
      }
      else
      if (strsame (cptr, "rss=", 4))
      {
         zptr = (sptr = KeyWord) + sizeof(KeyWord)-1;
         for (cptr += 4;
              *cptr && *cptr != '$' && sptr < zptr;
              *sptr++ = *cptr++);
         *sptr = '\0';
         CgiLibHtmlDeEntify(KeyWord);
         ThisResult = strsame (ContentRssFeedNamePtr, KeyWord, -1);
      }
      else
      if (strsame (cptr, "rss", -1) ||
          strsame (cptr, "rss$", -1))
         ThisResult = QueryRss;
      else
      if (strsame (cptr, "search", -1) ||
          strsame (cptr, "search$", -1))
         ThisResult = QueryKeyword;
      else
         TemplateParamError ("IF", SubsArg, cptr);

      if (WithNot) ThisResult = !ThisResult;
      if (WithAnd)
         TestResult = TestResult && ThisResult;
      else
         TestResult = TestResult || ThisResult;

      while (*cptr && *cptr != '$') cptr++;
      if (*cptr) cptr++;

      if (WatchEnabled)
         WatchThis (FI_LI, "TEST so-far:!AZ this:!AZ \"!AZ\"",
                    TestResult ? "TRUE" : "FALSE",
                    ThisResult ? "TRUE" : "FALSE", ParamPtr);
   }

   if (WatchEnabled)
      WatchThis (FI_LI, "TEST result:!AZ",
                 TestResult ? "TRUE" : "FALSE");

   return (TestResult);
}

/*****************************************************************************/
/*
This function includes/inserts text from outside of the current content.

The basic difference between including and inserting content, is an inclusion
is obvious, with a border and scrollbars as necessary (implemented as an
IFRAME).  With an insertion VWcms places the content in-line to the main
content of the page (implemented as a DIV).  It is much less obvious that it
has been retrieved from elsewhere.

If the target specified does not begin with as an obvious URI (e.g. with
http://, https://, html:///, plain:///, /, etc.) then the content is included
directly into the page (i.e. as unescaped HTML, etc.)  If the target begins
with a '/' then it's considered a local URL and included in an IFRAME as are
identifiable URLs beginning http://, https://, etc.

If the target begins html:// then the content is included as-is (careful, can
distrupt page layout) and if plain:// then the content is HTML-escaped before
inclusion.

If the target begines vwcms:// then the content of another section of the same
site is included/inserted.  Such vwcms:// sections cannot be nested.

If enabled, spawn:/// (HTML output) and splain:/// (HTML-escaped output) spawn
a subprocess and include/insert the output of that.
*/

void TemplateInclude
(
BOOL InsertInto,
char *SubsArg
)
{
   static char  DefaultIncludeDivStyle [] =
"border:1px dotted gray;overflow:auto;height:100%;width:100%;";

   static char  DefaultIncludeIframeStyle [] =
"border:1px dotted gray;overflow:auto;height:100%;width:100%;";

   static char  DefaultInsertIframeStyle [] =
"border-width:0;border-style:none;height:100%;width:100%;overflow:hidden;";

   static char  DefaultInsertDivStyle [] =
"border-width:0;border-style:none;height:100%;width:100%;overflow:hidden;";

   BOOL  FileHtml,
         FilePlain,
         SpawnHtml,
         SpawnPlain,
         VWcmsSection;
   int  status,
        RefreshSeconds = 0,
        TextLength;
   char  *cptr, *sptr, *zptr,
         *ClassPtr,
         *DefaultDivStylePtr,
         *DefaultIframeStylePtr,
         *OutputPtr = NULL,
         *SourcePtr = "",
         *TextPtr;
   char  FileName [256],
         HeightOf [64] = "",
         OnLoad [256] = "",
         RefreshAt [64] = "",
         SourceBuffer [256] = "",
         StyleBuffer [256] = "",
         WidthOf [64] = "";
   CONTENT_DATA  *cdptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "TemplateInclude() %d |%s|\n", InsertInto, SubsArg);

   /* awkward but backwardly compatible */
   if (*(cptr = SubsArg) == '$') cptr++;
   while (*cptr)
   {
      if (strsame (cptr, "height=", 7))
         cptr = TemplateParseParam (cptr+7, HeightOf, sizeof(HeightOf));
      else
      if (strsame (cptr, "refresh=", 8))
      {
         cptr = TemplateParseParam (cptr+8, RefreshAt, sizeof(RefreshAt));
         RefreshSeconds = atoi(RefreshAt);
      }
      else
      if (strsame (cptr, "source=", 7))
         cptr = TemplateParseParam (cptr+7, SourcePtr = SourceBuffer,
                                    sizeof(SourceBuffer));
      else
      if (strsame (cptr, "style=", 6))
         cptr = TemplateParseParam (cptr+6, StyleBuffer, sizeof(StyleBuffer));
      else
      if (strsame (cptr, "width=", 6))
         cptr = TemplateParseParam (cptr+6, WidthOf, sizeof(WidthOf));
      else
         break;
   }
   if (HeightOf[0] || RefreshAt[0] || SourceBuffer[0] ||
       StyleBuffer[0] || WidthOf[0])
   {
      if (*cptr)
      {
         if (InsertInto)
            TemplateParamError ("INSERT", SubsArg, cptr);
         else
            TemplateParamError ("INCLUDE", SubsArg, cptr);
         return;
      }
   }
   else
   {
      /* if none of the above have been supplied fall back to old habits */
      SourcePtr = cptr;
      if (strchr (cptr, '$'))
      {
         zptr = (sptr = StyleBuffer) + sizeof(StyleBuffer)-1;
         while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
         *sptr = '\0';
         if (*cptr == '$') cptr++;
         SourcePtr = cptr;
      }
   }

   if (!*SourcePtr)
   {
      if (InsertInto)
         TemplateParamError ("INSERT", NULL, NULL);
      else
         TemplateParamError ("INCLUDE", NULL, NULL);
      return;
   }

   FileHtml = FilePlain = SpawnHtml = SpawnPlain = VWcmsSection = FALSE;
   if (!strncmp (SourcePtr, "html:///", 8))
   {
      FileHtml = TRUE;
      SourcePtr += 8;
   }
   else
   if (!strncmp (SourcePtr, "plain:///", 9))
   {
      FilePlain = TRUE;
      SourcePtr += 9;
   }
   else
   if (!strncmp (SourcePtr, "spawn:///", 9))
   {
      SpawnHtml = TRUE;
      SourcePtr += 9;
   }
   else
   if (!strncmp (SourcePtr, "splain:///", 10))
   {
      SpawnPlain = TRUE;
      SourcePtr += 10;
   }
   else
   if (!strncmp (SourcePtr, "vwcms:///", 9))
   {
      if (TemplateInsertContentCount)
      {
         if (InsertInto)
            TemplateReportError ("INSERT", SourcePtr, "cannot nest vwcms:///");
         else
            TemplateReportError ("INCLUDE", SourcePtr, "cannot nest vwcms:///");
         return;
      }
      VWcmsSection = TRUE;
   }
   else
   {
      /* a URL? */
      for (sptr = SourcePtr; *sptr && *sptr != '/'; sptr++);
      if (!*sptr) FilePlain = TRUE;
   }

   if (FileHtml || FilePlain)
   {
      zptr = (sptr = FileName) + sizeof(FileName)-1;
      /* look for a colon, default to the site directory if none */
      for (cptr = SourcePtr; *cptr && *cptr != ':'; cptr++);
      if (!*cptr)
         for (cptr = VWcmsConfig.SiteDirectory;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
      for (cptr = SourcePtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
      *sptr = '\0';

      status = DataReadFileIntoMemory (FileName, &TextPtr, &TextLength, TRUE);
      if (VMSok (status))
         if (FilePlain) TextPtr = HtmlEscape(TextPtr);
   }

   /* ensure it's all contained within the content width */
   if (!TemplateIncludeCount++)
      fprintf (stdout,
"<div id=\"include_0\" style=\"height:0;width:100%%;\"></div>\n\
<script type=\"text/javascript\">\n\
if (!VWcmsIncludeWidth) {\n\
   var incobj = document.getElementById(\"include_0\");\n\
   if (incobj && incobj.clientWidth) \
var VWcmsIncludeWidth = incobj.clientWidth + \"px\";\n\
   if (!VWcmsIncludeWidth) \
var VWcmsIncludeWidth = Math.round(document.body.scrollWidth * %d / 100) \
+ \"px\";\n\
}\n\
function adjustIncludeWxH(number,width,height) {\n\
   var incobj = document.getElementById(\"include_\"+number);\n\
   if (width) incobj.style.width = width;\n\
   if (height) incobj.style.height = height;\n\
}\n\
function adjustIncludeW(number) {\n\
   if (VWcmsIncludeWidth) document.getElementById(\"include_\"+number)\
.style.width = VWcmsIncludeWidth;\n\
}\n\
function resizeInclude (number,bodyObj) {\n\
// Adjust the vertical height of the iframe to the inner document.\n\
// Can be a dynamic document (script) changing height.\n\
// I'm happy to be shown a better method that this!\n\
   var iFrameObj = document.getElementById(\"include_\"+number);\n\
   iFrameObj.scrolling = \"no\";\n\
   iFrameObj.style.overflow = \"hidden\";\n\
   iFrameObj.style.height = bodyObj.offsetHeight + \"px\";\n\
   iFrameObj.style.height = bodyObj.scrollHeight + \"px\";\n\
   iFrameObj.style.width = bodyObj.scrollWidth + \"px\";\n\
}\n\
</script>\n",
               PragmaMsieIncludePercent);

   if (InsertInto)
   {
      ClassPtr = "insert";
      DefaultDivStylePtr = DefaultInsertDivStyle;
      DefaultIframeStylePtr = DefaultInsertIframeStyle;
   }
   else
   {
      ClassPtr = "include";
      DefaultDivStylePtr = DefaultIncludeDivStyle;
      DefaultIframeStylePtr = DefaultIncludeIframeStyle;
   }

   if (SpawnHtml || SpawnPlain)
   {
      /* (effectively) 'frame' the subprocess output */
      if (SpawnPlain)
         fprintf (stdout,
"<div id=\"include_%d\" class=\"vwcms_%s\" \
style=\"white-space:pre;%s%s%s\">",
                  TemplateIncludeCount,
                  ClassPtr,
                  FONT_FAMILY_MONOSPACE,
                  DefaultDivStylePtr, StyleBuffer);
      else
         fprintf (stdout,
"<div id=\"include_%d\" class=\"vwcms_%s\" style=\"%s%s\">",
                  TemplateIncludeCount,
                  ClassPtr,
                  DefaultDivStylePtr, StyleBuffer);

      if (VMSnok (status = SpawnCommand (SourcePtr, &OutputPtr)))
         fprintf (stdout,
"<span class=\"vwcms_include_error\">%s (%%X%08.08X)</span>",
                  SysGetMsg(status), status);

      if (OutputPtr)
         if (SpawnPlain)
            fputs (HtmlEscape(OutputPtr), stdout);
         else
            fputs (OutputPtr, stdout);

      fprintf (stdout, "</div>\n");
   }
   else
   if (FileHtml || FilePlain)
   {
      /* (effectively) 'frame' a file */
      if (FilePlain)
         fprintf (stdout,
"<div id=\"include_%d\" class=\"vwcms_%s\" \
style=\"white-space:pre;%s%s%s\">",
                  TemplateIncludeCount,
                  ClassPtr,
                  FONT_FAMILY_MONOSPACE,
                  DefaultDivStylePtr, StyleBuffer);
      else
         fprintf (stdout,
"<div id=\"include_%d\" class=\"vwcms_%s\" style=\"%s%s\">",
                  TemplateIncludeCount,
                  ClassPtr,
                  DefaultDivStylePtr, StyleBuffer);

      if (VMSok(status))
         fputs (TextPtr, stdout);
      else
         fprintf (stdout,
"<span class=\"vwcms_include_error\">%s (%%X%08.08X)</span>",
                  SysGetMsg(status), status);

      fprintf (stdout, "</div>\n");
   }
   else
   if (VWcmsSection)
   {
      if (!(cdptr = ContentFindSection (SourcePtr+9)))
      {
         TemplateReportError ("INCLUDE", SourcePtr, "not found");
         return;
      }
      fprintf (stdout,
"<div id=\"include_%d\" class=\"vwcms_%s\" style=\"%s%s\">",
               TemplateIncludeCount,
               ClassPtr,
               DefaultDivStylePtr, StyleBuffer);
      TemplateInsertContentCount++;
      TemplateContent (cdptr);
      TemplateInsertContentCount--;
      fprintf (stdout, "</div>\n");
   }
   else
   {
      /* put a URL into a frame */
      sptr = OnLoad;
      if (WidthOf[0] || HeightOf[0])
      {
         sptr += sprintf (sptr, " onload=\"");
         if (WidthOf[0])
            sptr += sprintf (sptr, "this.style.width=\'%s\';", WidthOf);
         if (HeightOf[0])
            sptr += sprintf (sptr, "this.style.height=\'%s\';", HeightOf);
      }
      else
      {
         /* ensure the content fills the iframe */
         sptr += sprintf (sptr,
"onload=\"resizeInclude(%d,frames[\'include_%d\'].document.body);",
                  TemplateIncludeCount, TemplateIncludeCount);
      }
      if (RefreshSeconds)
         sprintf (sptr,
"includeReload%d=\'include_%d\';\
setTimeout(\'frames[includeReload%d].document.location.reload(true)\',%d);\"",
                  TemplateIncludeCount, TemplateIncludeCount,
                  TemplateIncludeCount, RefreshSeconds*1000);
      else
         strcpy (sptr, "\"");

      fprintf (stdout,
"<iframe id=\"include_%d\" name=\"include_%d\" \
frameborder=\"0\" scrolling=\"auto\"%s \
class=\"vwcms_%s\" style=\"%s%s\" \
src=\"%s\"></iframe>\n",
               TemplateIncludeCount, TemplateIncludeCount,
               OnLoad,
               ClassPtr,
               DefaultIframeStylePtr, StyleBuffer, SourcePtr);
   }

   if (WidthOf[0] || HeightOf[0])
   {
      if (FileHtml || FilePlain ||
          SpawnHtml || SpawnPlain)
      {
         /* ensure it's set to request width and height */
         fprintf (stdout,
"<script type=\"text/javascript\">\
setTimeout (\"adjustIncludeWxH(%d,%s%s%s,%s%s%s)",
                  TemplateIncludeCount,
                  WidthOf[0] ? "\'" : "",
                  WidthOf[0] ? WidthOf : "null",
                  WidthOf[0] ? "\'" : "",
                  HeightOf[0] ? "\'" : "",
                  HeightOf[0] ? HeightOf : "null",
                  HeightOf[0] ? "\'" : "");
         if (FilePlain)
            fprintf (stdout, ";adjustIncludeW(%d)", TemplateIncludeCount);
         fputs ("\",50)</script>\n", stdout);
      }
   }
   else
   {
      /* ensure it's all contained within the content width */
      fprintf (stdout,
"<script type=\"text/javascript\">\
setTimeout (\"adjustIncludeW(%d)\",50);\
</script>\n",
               TemplateIncludeCount);
   }
}

/*****************************************************************************/
/*
Insert the content of the section specified by $$CONTENT$The/Content$$. 
Parallel functionality is provided by $$INSERT$vwcms:///The/Content$$ and
with a different format by $$INCLUDE$vwcms:///The/Content$$.
*/

void TemplateInsertContent (char *SubsArg)

{
   char  *cptr;
   CONTENT_DATA  *cdptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateInsertContent() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;
   while (*cptr && ISLWS(*cptr)) cptr++;
   if (*cptr == '/') cptr++;

   if (!*cptr)
      TemplateParamError ("CONTENT", NULL, NULL);
   else
   if (!(cdptr = ContentFindSection (cptr)))
      TemplateReportError ("CONTENT", cptr, "not found");
   else
   {
      TemplateInsertContentCount++;
      TemplateContent (cdptr);
      TemplateInsertContentCount--;
   }
}

/*****************************************************************************/
/*
The content last-modified date/time.

If the first parameter is '#' the provide the section last-modified.
If empty or '*' then provide the document last-modified.

If no second parameter provide a generic date/time with day of week.  If the
parameter is "VMS", "VMS20", "VMS17" or "VMS11" generate a VMS-style time
string, terminated optionally at the hundreds, minutes or date respectively. 
The parameter string can also provide a strftime() format string.
*/

void TemplateLastModified (char* SubsArg)

{
   char  *cptr, *sptr;
   ulong  LastUpdated;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateLastModified() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;
   sptr = cptr;
   for (sptr = cptr; *sptr && *sptr != '$'; sptr++);
   if (*sptr) sptr++;

   if (*cptr && *cptr == '#') {
      if (!(LastUpdated = SectionLastUpdated))
         LastUpdated = ContentLastUpdated; 
   }
   else
      LastUpdated = ContentLastUpdated; 

   sptr = TimeString (sptr, LastUpdated);

   if (*sptr == '!')
      TemplateReportError ("LASTMOD", SubsArg, sptr+1);
   else
      fputs (sptr, stdout);
}

/*****************************************************************************/
/*
If the user is currently logged in then provide a link to allow logout.  If not
logged-in and not currently in the process of loggin-in ("i.e. ?login"),
provide a link to request login credentials.

Three parameters can be passed to this.  The first is the alternate text for
the "Login" link; the second the alternate text for the "Logout" link; and the
third, if present, is the separator between the "Logout" and the logged-in
username.
*/

void TemplateLoginLogout (char* SubsArg)

{
   char  *cptr, *sptr, *zptr,
         *LoginPtr,
         *LogoutPtr,
         *SectionNamePtr,
         *WebServerPtr;
   char  LoginString [256],
         LogoutString [256],
         WebServerString [256];
   CONTENT_DATA  *SectionPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateLoginLogout() |%s|\n", SubsArg);

   /* if can be but not currently web server authenticated */
   if (VWcmsConfig.LoginRemoteUser && !LoginUserName[0]) return;

   if (*(cptr = SubsArg) == '$') cptr++;

   cptr = TemplateParseParam (cptr, LoginString, sizeof(LoginString));
   if (!*(LoginPtr = LoginString))
      LoginPtr = CONFIG_LANG_PARSE(LangLogin,1);

   cptr = TemplateParseParam (cptr, LogoutString, sizeof(LogoutString));
   if (!*(LogoutPtr = LogoutString))
      LogoutPtr = CONFIG_LANG_PARSE(LangLogin,2);

   cptr = TemplateParseParam (cptr, WebServerString, sizeof(WebServerString));
   if (!*(WebServerPtr = WebServerString))
      WebServerPtr = CONFIG_LANG_PARSE(LangLogin,3);

   if (ContentSectionPtr)
      SectionNamePtr = ContentSectionPtr->SectionName;
   else
      SectionNamePtr = ContentSectionName;

   if (VWcmsConfig.LoginRemoteUser)
   {
      /*****************************/
      /* web server authentication */
      /*****************************/

      if (cptr = IsAnImage(WebServerPtr))
         fprintf (stdout,
"<span class=\"vwcms_login_remote_img\"><img alt=\"[login]\" %s />%s</span>",
                  cptr, HtmlEscape(LoginUserName));
      else
         fprintf (stdout,
"<span class=\"vwcms_login_remote_img\">%s:&nbsp;%s</span>",
                  WebServerPtr, HtmlEscape(LoginUserName));
   }
   else
   if (!LoginUserName[0] && !QueryLogin)
   {
      /***************************/
      /* currently not logged-in */
      /***************************/

      if (cptr = IsAnImage(LoginPtr))
         fprintf (stdout,
"<span class=\"vwcms_login_img\"><a%s href=\"%s?login\">\
<img alt=\"[LOGIN]\" %s /></a></span>",
                  VWcmsConfig.LoginEnabled ? "" : "_disabled",
                  ContentEncodeName("/",SectionNamePtr),
                  cptr);
      else
         fprintf (stdout,
"<span class=\"vwcms_login\"><a%s href=\"%s?login\">%s</a></span>",
                  VWcmsConfig.LoginEnabled ? "" : "_disabled",
                  ContentEncodeName("/",SectionNamePtr),
                  LoginPtr);
   }
   else
   if (LoginUserName[0] && !QueryLogout)
   {
      /***********************/
      /* currently logged-in */
      /***********************/

      if (cptr = IsAnImage(LogoutPtr))
         fprintf (stdout,
"<span class=\"vwcms_logout_img\"><a%s href=\"%s?logout\">\
<img alt=\"[LOGOUT]\" %s />%s\
<span id=\"login_idle_reminder\"></span></a></span>",
                  ManageLeaveNow(),
                  ContentEncodeName("/",SectionNamePtr),
                  cptr, HtmlEscape(LoginUserName));
      else
         fprintf (stdout,
"<span class=\"vwcms_logout\"><a%s href=\"%s?logout\">\
%s:&nbsp;%s<span id=\"login_idle_reminder\"></span></a></span>",
                  ManageLeaveNow(),
                  ContentEncodeName("/",SectionNamePtr),
                  LogoutPtr, HtmlEscape(LoginUserName));

      if (VWcmsConfig.LoginIdleSeconds)
      {
         fprintf (stdout,
"<script type=\"text/javascript\">\n\
var loginIdleSecs = %d;\n\
function loginIdleTick() {\n\
   if (loginIdleSecs < 120 || loginIdleSecs %% 10 == 0) {\n",
                  VWcmsConfig.LoginIdleSeconds);

         if (VWcmsConfig.LoginIdleSeconds >= 3600)
            fputs (
"      var hh = Math.floor(loginIdleSecs/3600);\n\
      if (hh == 0) \n\
         hh = \"\";\n\
      else \n\
      if (hh <= 9) hh = \"0\" + hh;\n\
      if (hh != \"\") hh = hh + \":\";\n",
                   stdout);
         else
            fputs ("      hh = \"\";\n", stdout);

         fputs (
"      var mm = Math.floor((loginIdleSecs%3600)/60);\n\
      if (mm <= 9) mm = \"0\" + mm;\n\
      mm = mm + \":\";\n\
      var ss = Math.floor(loginIdleSecs%60);\n\
      if (ss <= 9) ss = \"0\" + ss;\n\
      document.getElementById(\"login_idle_reminder\").innerHTML = \
\"&nbsp;\"+hh+mm+ss;\n\
   }\n\
   if (loginIdleSecs == 0) return;\n\
   loginIdleSecs--;\n\
   setTimeout(\"loginIdleTick()\",1000);\n\
}\n\
loginIdleTick();\n\
</script>\n",
                  stdout);
      }

      TemplateLogout = TRUE;
   }
}

/*****************************************************************************/
/*
Macro from the configuration file with optional string parameters.
*/

void TemplateMacro (char *SubsArg)

{
   char  *cptr, *mptr, *sptr, *zptr,
         *MacroPtr;
   char  MacroName [256],
         StrBuf [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateMacro() |%s|\n", SubsArg);

   cptr = TemplateParseParam (SubsArg, MacroName, sizeof(MacroName));

   if (!(mptr = ConfigGetMacro (MacroName+1)))
   {
      /* returns a pointer to the macro *value* */
      TemplateReportError (SubsArg, NULL, CONFIG_LANG_PARSE(LangSubsError,3));
      return;
   }

   if (*mptr++ != ':')
   {
      /* if it's not a substitution macro */
      TemplateReportError (SubsArg, NULL, CONFIG_LANG_PARSE(LangSubsError,3));
      return;
   }

   for (;;)
   {
      MacroPtr = sptr = mptr;
      while (*mptr)
      {
         if (*(USHORTPTR)mptr == '%s') break;
         if (*(USHORTPTR)mptr == '%%') break;
         mptr++;
      }
      OutputFromTo (sptr, mptr);
      if (*(USHORTPTR)mptr == '%%')
      {
         mptr++;
         continue;
      }
      if (!*mptr) break;

      if (!*cptr)
      {
         TemplateParamError (mptr, MacroName, MacroName);
         break;
      }

      /* must have hit a '%s' string */
      mptr += 2;

      zptr = (sptr = StrBuf) + sizeof(StrBuf)-1;
      while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';
      while (*cptr && *cptr != '$') cptr++;
      if (*cptr) cptr++;

      if (StrBuf[0]) fputs (StrBuf, stdout);
   }

   if (*cptr) TemplateParamError (cptr, MacroName, MacroName);
}

/*****************************************************************************/
/*
If the management interface is currently visible then provide a link to change
to the site view, if viewing the site then a link to the management interface.
*/

void TemplateManage (char* SubsArg)

{
   char  *cptr, *sptr, *zptr,
         *SectionNamePtr,
         *ManagePtr;
   char  ManageString [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateManage() |%s|\n", SubsArg);

   if (!ManageSiteOn) return;

   /* bit of a kludge but for historical reasons */
   ManageAbsoluteX = ManageAbsoluteY = 0;
   if (*SubsArg &&
       (cptr = strstr (SubsArg, "absolute=")) ||
       (cptr = strstr (SubsArg, "ABSOLUTE=")))
   {
      *cptr = '\0';
      cptr += 9;
      ManageAbsoluteX = atoi(cptr);
      while (*cptr && isdigit(*cptr)) cptr++;
      while (*cptr && !isdigit(*cptr)) cptr++;
      ManageAbsoluteY = atoi(cptr);
   }

   if (*(cptr = SubsArg) == '$') cptr++;
   if (ManageEditOn)
   {
      /* skip any first parameter (e.g. "Manage") */
      while (*cptr && *cptr != '$') cptr++;
      if (*cptr) cptr++;
      /* now at any second parameter (e.g. "View") */
   }
   zptr = (sptr = ManageString) + sizeof(ManageString)-1;
   while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';

   if (!*(ManagePtr = ManageString))
      if (ManageEditOn)
         ManagePtr = CONFIG_LANG_PARSE(LangManage,2);
      else
         ManagePtr = CONFIG_LANG_PARSE(LangManage,1);

   if (ContentSectionPtr)
      SectionNamePtr = ContentSectionPtr->SectionName;
   else
      SectionNamePtr = ContentSectionName;

   if (!ManageEditOn)
      if (cptr = IsAnImage (ManagePtr))
         fprintf (stdout,
"<span class=\"vwcms_manage_img\"><a href=\"%s?manage\">\
<img alt=\"[MANAGE]\" %s /></a></span>",
                  ContentEncodeName("/",SectionNamePtr),
                  cptr);
      else
         fprintf (stdout,
"<span class=\"vwcms_manage\"><a href=\"%s?manage\">%s</a></span>",
                  ContentEncodeName("/",SectionNamePtr),
                  ManagePtr);
   else
      if (cptr = IsAnImage (ManagePtr))
         fprintf (stdout,
"<span class=\"vwcms_manage_img\"><a href=\"%s?view\">\
<img alt=\"[VIEW]\" %s /></a></span>",
                  ContentEncodeName("/",SectionNamePtr),
                  cptr);
      else
         fprintf (stdout,
"<span class=\"vwcms_manage\"><a%s href=\"%s?view\">%s</a></span>",
                  ManageLeaveNow(),
                  ContentEncodeName("/",SectionNamePtr),
                  ManagePtr);
}

/*****************************************************************************/
/*
*/

void TemplateNext (char* SubsArg)

{
   char  *cptr, *sptr, *zptr,
         *NextPtr,
         *SectionNamePtr;
   char  NextString [256];
   CONTENT_DATA  *cdptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateNext() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;
   zptr = (sptr = NextString) + sizeof(NextString)-1;
   while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';
   if (!*(NextPtr = NextString))
      NextPtr = CONFIG_LANG_PARSE(LangNavigate,2);

   if (ContentSectionPtr)
   {
      for (cdptr = ContentSectionPtr->NextPtr; cdptr; cdptr = cdptr->NextPtr)
         if (!cdptr->IsHidden) break;
      if (cdptr)
         SectionNamePtr = cdptr->SectionName;
      else
         SectionNamePtr = NULL;
   }
   else
      SectionNamePtr = NULL;

   if (cptr = IsAnImage(NextPtr))
   {
      if (SectionNamePtr)
         fprintf (stdout,
"<a%s href=\"%s\"><img class=\"vwcms_nav_img\" alt=\"[NEXT]\" %s /></a>",
                  ManageLeaveNow(),
                  ContentEncodeName("/",SectionNamePtr),
                  cptr);
      else
         fprintf (stdout,
"<img class=\"vwcms_nav_img_off\" alt=\"[NEXT]\" %s />", cptr);
   }
   else
   {
      if (SectionNamePtr)
         fprintf (stdout,
"<a%s href=\"%s\"><span class=\"vwcms_nav\">%s</span></a>",
                  ManageLeaveNow(),
                  ContentEncodeName("/",SectionNamePtr),
                  NextPtr);
      else
         fprintf (stdout, "<span class=\"vwcms_nav\">%s</span>", NextPtr);
   }
}

/*****************************************************************************/
/*
A hodge-podge of directive(s).
*/

void TemplatePragma (char* SubsArg)

{
   char  *cptr, *sptr, *zptr;
   char  PragmaName [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplatePragma() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;
   zptr = (sptr = PragmaName) + sizeof(PragmaName)-1;
   while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';
   if (*cptr) cptr++;

   if (strsame (PragmaName, "MSIE_INCLUDE_PERCENT", -1))
   {
      PragmaMsieIncludePercent = atoi(cptr);
      /* bit of a sanity check */
      if (PragmaMsieIncludePercent < 10)
         PragmaMsieIncludePercent = 100;
      else
      if (PragmaMsieIncludePercent > 200)
         PragmaMsieIncludePercent = 100;
   }
   else
      TemplateReportError ("PRAGMA", SubsArg, PragmaName);
}

/*****************************************************************************/
/*
*/

void TemplatePrev (char* SubsArg)

{
   char  *cptr, *sptr, *zptr,
         *PrevPtr,
         *SectionNamePtr;
   char  PrevString [256];
   CONTENT_DATA  *cdptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplatePrev() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;
   zptr = (sptr = PrevString) + sizeof(PrevString)-1;
   while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';
   if (!*(PrevPtr = PrevString))
      PrevPtr = CONFIG_LANG_PARSE(LangNavigate,1);

   if (ContentSectionPtr)
   {
      for (cdptr = ContentSectionPtr->PrevPtr; cdptr; cdptr = cdptr->PrevPtr)
         if (!cdptr->IsHidden) break;
      if (cdptr)
         SectionNamePtr = cdptr->SectionName;
      else
         SectionNamePtr = NULL;
   }
   else
      SectionNamePtr = NULL;

   if (cptr = IsAnImage(PrevPtr))
   {
      if (SectionNamePtr)
         fprintf (stdout,
"<a%s href=\"%s\"><img class=\"vwcms_nav_img\" alt=\"[PREV]\" %s /></a>",
                  ManageLeaveNow(),
                  ContentEncodeName("/",SectionNamePtr),
                  cptr);
      else
         fprintf (stdout,
"<img class=\"vwcms_nav_img_off\" alt=\"[PREV]\" %s />", cptr);
   }
   else
   {
      if (SectionNamePtr)
         fprintf (stdout,
"<a%s href=\"%s\"><span class=\"vwcms_nav\">%s</span></a>",
                  ManageLeaveNow(),
                  ContentEncodeName("/",SectionNamePtr),
                  PrevPtr);
      else
         fprintf (stdout, "<span class=\"vwcms_nav\">%s</span>", PrevPtr);
   }
}

/*****************************************************************************/
/*
If 'SubsArg' begins with a digit this represents the range of print to be
requested.  0 is the current section (default anyway but with a section number
prefixing the section title), 1..6 the current and subordinate sections, and >6
the entire site.
*/

void TemplatePrint (char* SubsArg)

{
   int  PrintRange = 0;
   char  *cptr, *sptr, *zptr,
         *PrintPtr,
         *NoPrintDialog,
         *QueryPtr,
         *SectionNamePtr;
   char  PrintString [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplatePrint() |%s|\n", SubsArg);

   if (ContentSectionPtr)
      SectionNamePtr = ContentSectionPtr->SectionName;
   else
      SectionNamePtr = ContentSectionName;

   if (*(cptr = SubsArg) == '$') cptr++;

   if (*cptr == '_')
   {
      NoPrintDialog = "_";
      cptr++;
   }
   else
      NoPrintDialog = "";

   if (isdigit(*cptr))
   {
      PrintRange = atoi(cptr);
      while (isdigit(*cptr)) cptr++;
      if (*cptr == '$') cptr++;
   }
   else
   if (*cptr == '*')
   {
      PrintRange = 9;
      if (*++cptr == '$') cptr++;
   }
   else
   if (*cptr == '+')
   {
      PrintRange = 1;
      if (*++cptr == '$') cptr++;
   }
   else
   if (NoPrintDialog)
   {
      /* if something like $$PRINT$_$Print Page$$ */
      if (*cptr == '$') cptr++;
   }

   zptr = (sptr = PrintString) + sizeof(PrintString)-1;
   while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';

   if (!*(PrintPtr = PrintString))
   {
      if (PrintRange == 0)
         PrintPtr = CONFIG_LANG_PARSE(LangPrint,1);
      else
      if (PrintRange <= 6)
         PrintPtr = CONFIG_LANG_PARSE(LangPrint,3);
      else
         PrintPtr = CONFIG_LANG_PARSE(LangPrint,5);
   }

   if (PrintRange == 9)
      QueryPtr = CONFIG_LANG_PARSE(LangPrint,6);
   else
   if (ContentSectionPtr &&
       ContentSectionPtr->TextLength)
   {
      /* this section has content */
      if (PrintRange >= 1)
      {
         /* check for subordinate sections */
         if (ContentSectionPtr->NextPtr &&
             ContentSectionPtr->NextPtr->TextLength &&
             ContentSectionPtr->NextPtr->HeadNumber >
                ContentSectionPtr->HeadNumber) 
            QueryPtr = CONFIG_LANG_PARSE(LangPrint,4);
         else
            QueryPtr = NULL;
      }
      else
         QueryPtr = CONFIG_LANG_PARSE(LangPrint,2);
   }
   else
      QueryPtr = NULL;

   if (cptr = IsAnImage(PrintPtr))
   {
      if (QueryPtr)
         fprintf (stdout,
"<a%s href=\"%s?%s%s\">\
<img class=\"vwcms_print_img\" alt=\"[PRINT]\" %s /></a>",
                  ManageLeaveNow(),
                  ContentEncodeName("/",SectionNamePtr),
                  NoPrintDialog, QueryPtr, cptr);
       else
         fprintf (stdout,
"<img class=\"vwcms_print_img_off\" alt=\"[PRINT]\" %s />", cptr);
   }
   else
   {
      if (QueryPtr)
         fprintf (stdout,
"<span class=\"vwcms_print\"><a%s href=\"%s?%s%s\">%s</a></span>",
                  ManageLeaveNow(),
                  ContentEncodeName("/",SectionNamePtr),
                  NoPrintDialog, QueryPtr, PrintPtr);
      else
         fprintf (stdout, "<span class=\"vwcms_print\">%s</span>", PrintPtr);
   }
}

/*****************************************************************************/
/*
Output a range of sections.  If the argument 'HeadNumber' is zero then output
just the current section.  If between one and six then output the
'ContentPtr'ed-to section and all subordinate sections.  If greater than six
then output the entire site.
*/

void TemplateRange (int HeadNumber)

{
   int  NoPrintHead;
   char  *cptr;
   char  ItemBuffer [128];
   CONTENT_DATA  *cdptr,
                 *ContentPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateRange() %d\n", HeadNumber);

   if (!ContentSitePtr || !ContentSectionPtr) return;

   if (HeadNumber == 0)
   {
      /* just this section */
      ContentPtr = ContentSectionPtr;
   }
   else
   if (HeadNumber > 0 && HeadNumber <= 6)
   {
      /* this section and all subordinate sections */
      ContentPtr = ContentSectionPtr;
      HeadNumber = ContentPtr->HeadNumber;
   }
   else
      ContentPtr = ContentSitePtr;

   for (cdptr = ContentPtr; cdptr; cdptr = cdptr->NextPtr)
   {
      if (cdptr->IsHidden) continue;
      if (QueryPrint && cdptr->NoPrint) continue;
      if (QueryPrint && cdptr->NoPrintPlus)
      {
         /* absorb this and any subordinates */
         NoPrintHead = cdptr->HeadNumber;
         while (cdptr->NextPtr)
         {
            if (cdptr->NextPtr->HeadNumber <= NoPrintHead) break;
            cdptr = cdptr->NextPtr;
         }
         continue;
      }

      /* after the first section check if we're finished the range */
      if (cdptr != ContentPtr &&
          (HeadNumber == 0 ||
           (HeadNumber <= 6 && HeadNumber >= cdptr->HeadNumber)))
         break;

      /* give the item an 'ordered list' type number */
      if (HeadNumber)
         fprintf (stdout, "\n<h%d>%s&nbsp;&nbsp;%s</h%d>\n\n",
                  cdptr->HeadNumber,
                  cdptr->SectionNumberDot,
                  HtmlEscape(cdptr->SectionTitle),
                  cdptr->HeadNumber);
      else
         fprintf (stdout, "\n<h1>%s</h1>\n\n",
                  HtmlEscape(cdptr->SectionTitle));

      /* adjust the global section pointer ffor prints, etc. */
      TemplateContent (ContentSectionPtr = cdptr);
   }
}

/*****************************************************************************/
/*
*/

void TemplateRotate (char* SubsArg)

{
   static int  ClassNumber;

   int  deg;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateRotate() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;

   if (*cptr == '-' || isdigit(*cptr))
   {
      deg = atoi(cptr);
      fprintf (stdout,
"<span style=\" \
-o-transform: rotate(%ddeg); \
-webkit-transform: rotate(%ddeg); \
-moz-transform: rotate(%ddeg); \
\">",
               deg, deg, deg);
   }
   else
      fputs ("</span>", stdout);
}

/*****************************************************************************/
/*
Really Simple Syndication.
Two purposes:
1) In-text 'subscribe to' link.
2) In-text 'validate my feed' link.
*/

void TemplateRSS (char* SubsArg)

{
   BOOL  ValidateFeed = FALSE,
         SubscribeTo = FALSE;
   int  len;
   char  *aptr, *cptr, *sptr, *zptr,
         *SubscribePtr,
         *ValidatePtr,
         *ValidatorPtr;
   char  RssFeed [256] = "",
         RssFeedUrl [1024] = "",
         RssSubscribe [256] = "",
         RssValidate [256] = "",
         RssValidator [256] = "",
         ValidatorUrl [256] = "";

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateRSS() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;

   while (*cptr)
   {
      if (strsame (cptr, "feed=", 5))
         cptr = TemplateParseParam (cptr+5, RssFeed, sizeof(RssFeed));
      else
      if (strsame (cptr, "subscribe[=]", 9))
      {
         SubscribeTo = TRUE;
         cptr = TemplateParseParam (cptr+9, RssSubscribe, sizeof(RssSubscribe));
      }
      else
      if (strsame (cptr, "validate[=]", 8))
      {
         ValidateFeed = TRUE;
         cptr = TemplateParseParam (cptr+8, RssValidate, sizeof(RssValidate));
      }
      else
      if (strsame (cptr, "validator=", 10))
      {
         ValidateFeed = TRUE;
         cptr = TemplateParseParam (cptr+10, RssValidator,
                                             sizeof(RssValidator));
      }
      else
         cptr = TemplateParseSkipParam (cptr, NULL);
   }

   if (!RssFeed[0])
   {
      CgiLibResponseHeader (200, "text/plain",
                            "Last-Modified: %s\n", LastModifiedNow);
      fprintf (stdout, "ERROR 500 (%s): $$RSS$feed=?$$\n",
               CgiLibHttpStatusCodeText(500));
      return;
   }

   if (SubscribeTo)
   {
      if (!*(SubscribePtr = RssSubscribe))
         SubscribePtr = CONFIG_LANG_PARSE(LangRss,1);

      if (cptr = IsAnImage (SubscribePtr))
         fprintf (stdout,
"<a%s href=\"%s%s?rss=%s\"><img class=\"vwcms_nav_img\" \
alt=\"[SUBSCRIBE]\" %s /></a>",
                  ManageLeaveNow(),
                  LinkSchemeHost,
                  ContentEncodeName("/",NULL),
                  ContentEncodeName(NULL,RssFeed),
                  cptr);
      else
         fprintf (stdout,
"[<a%s href=\"%s%s?rss=%s\">%s</a>]",
                  ManageLeaveNow(),
                  LinkSchemeHost,
                  ContentEncodeName("/",NULL),
                  ContentEncodeName(NULL,RssFeed),
                  SubscribePtr);
   }

   if (ValidateFeed)
   {
      if (!*(ValidatePtr = RssValidate))
         ValidatePtr = CONFIG_LANG_PARSE(LangRss,3);

      if (!*(ValidatorPtr = RssValidator))
         ValidatorPtr = RSS_VALIDATOR_DEFAULT;

      zptr = (sptr = ValidatorUrl) + sizeof(ValidatorUrl)-1;
      /* up to the first string substition */
      for (aptr = ValidatorPtr;
           *aptr && *(USHORTPTR)aptr != '%s' && sptr < zptr;
           *sptr++ = *aptr++);
      if (*aptr)
      {
         /* substitute the RSS feed URL */
         aptr += 2;
         for (cptr = LinkSchemeHost; *cptr && sptr < zptr; *sptr++ = *cptr++);
         for (cptr = ContentEncodeName("/",NULL);
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
         for (cptr = "?rss="; *cptr && sptr < zptr; *sptr++ = *cptr++);
         for (cptr = ContentEncodeName(NULL,RssFeed);
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
         /* finalise with anything remaining */
         while (*aptr && sptr < zptr) *sptr++ = *aptr++;
      }
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';

      if (cptr = IsAnImage (ValidatePtr))
         fprintf (stdout,
"<a%s href=\"%s\"><img class=\"vwcms_nav_img\" alt=\"[VALIDATE]\" %s /></a>",
                  ManageLeaveNow(),
                  ValidatorUrl,
                  cptr);
      else
         fprintf (stdout,
"[<a%s href=\"%s\">%s</a>]",
                  ManageLeaveNow(),
                  ValidatorUrl,
                  ValidatePtr);
   }
}

/*****************************************************************************/
/*
Style 1 (default) is is a text field followed by a button.
Style 2 is a text field with an equal length button below it.
*/

void TemplateSearch (char* SubsArg)

{
   BOOL  ImgSubmit = FALSE;
   int  TextWidth = 0;
   char  *cptr, *sptr,
         *KeywordPtr,
         *SearchPtr,
         *SectionNamePtr;
   char  SearchString [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateSearch() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;

   if (isdigit(*cptr)) cptr = TemplateParseRowCol (cptr, NULL, &TextWidth);
   if (!TextWidth) TextWidth = 15;

   TemplateParseParam (cptr, SearchString, sizeof(SearchString));

   if (!*(SearchPtr = SearchString))
      SearchPtr = CONFIG_LANG_PARSE(LangSearch,1);

   if (ContentSectionPtr)
      SectionNamePtr = ContentSectionPtr->SectionName;
   else
      SectionNamePtr = ContentSectionName;

   KeywordPtr = GetCgiVar ("FORM_KEYWORD");

   fprintf (stdout,
"<form class=\"vwcms_search\" method=\"get\" action=\"%s\">\
<input type=\"text\" size=\"%d\" name=\"keyword\" value=\"%s\" \
onfocus=\"this.select()\" />&nbsp;",
            ContentEncodeName("/",SectionNamePtr),
            TextWidth,
            HtmlEscape(KeywordPtr));

   if (cptr = IsAnImage(SearchPtr))
      fprintf (stdout,
"<input%s name=\"search\" type=\"image\" %s /></form>",
               ManageLeaveNow(),
               cptr);
   else
      fprintf (stdout,
"<input%s type=\"submit\" name=\"search\" value=\"%s\" /></form>",
               ManageLeaveNow(),
               SearchPtr);
}

/*****************************************************************************/
/*
*/

void TemplateShow (char* SubsArg)

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateShow() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;

   if (strsame (cptr, "LANG", -1))
      ConfigLangTable ();
   else
   if (strsame (cptr, "VERSION", -1))
      fputs (HtmlEscape(SoftwareId), stdout);
   else
   if (!ShowLocal (SubsArg))
      TemplateReportError ("SHOW", SubsArg, NULL);
}

/*****************************************************************************/
/*
LIB$STAT_TIMER() statistics.  If a parameter is supplied (empty is all that's
necessary) outputs in plain-sight, else inside an HTML comment.
*/

void TemplateStats (char* SubsArg)

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateStats() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;

   if (*cptr == '*')
      fputs (StatTimer(TRUE), stdout);
   else
      fprintf (stdout, "<!-- %s -->", StatTimer(TRUE));
}

/*****************************************************************************/
/*
Output the time.  If 'SecondsTime' is zero output the current time.
*/

void TemplateTime
(
char* SubsArg,
ulong SecondsTime
)
{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateTime() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;
   cptr = TimeString (cptr, SecondsTime);

   if (*cptr == '!')
      TemplateReportError ("TIME", SubsArg, cptr+1);
   else
      fputs (cptr, stdout);
}

/*****************************************************************************/
/*
Output the section title.

The parameter optional can indicate the depth of the title.  If 0 (deault) then
the title of the current section.  If 1 then the current <h1> title, 2 the
current <h2> title (if any), etc.  If +1 then the title of the parent, if
there is one.  For +2 then the parent of the parent, etc. (negative numbers are
synonyms).

It may have been explicitly set using $$TITLE$..$$ in the first line of the
section.  Ignore this instance if encountered (i.e. if it has a non-numeric
parameter).
*/

void TemplateTitle (char* SubsArg)

{
   int  TitleLevel = 0;
   char  *cptr, *sptr, *zptr,
         *TitlePtr = NULL;
   char  TitleBuffer [256];
   CONTENT_DATA  *cdptr = NULL;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateTitle() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;

   if (*cptr == '+' || *cptr == '-' || isdigit(*cptr))
   {
      TitleLevel = atoi(cptr);
      if (*cptr == '+') TitleLevel = (-TitleLevel);

      if (TitleLevel < 0)
      {
         /* i.e. if HeadNumber is 3 and TitleLevel = -1 then it ends up 2 */
         TitleLevel = ContentSectionPtr->HeadNumber + TitleLevel;
         if (!TitleLevel) return;
      }

      if (TitleLevel > VWcmsConfig.SiteDepth)
         TitleLevel = VWcmsConfig.SiteDepth;
      else
      if (TitleLevel < (-VWcmsConfig.SiteDepth)+1)
         TitleLevel = (-VWcmsConfig.SiteDepth)+1;

      while (*cptr && *cptr != '$') cptr++;
      if (*cptr) cptr++;
   }
   else
   if (*cptr)
   {
      /* first-line use of $$TITLE$$ to explicitly set title */
      return;
   }

   if (TitleLevel)
   {
      /* look backwards for the requested level title */
      for (cdptr = ContentSectionPtr->PrevPtr; cdptr; cdptr = cdptr->PrevPtr)
      {
         if (cdptr->IsHidden) continue;
         if (cdptr->HeadNumber == TitleLevel) break;
      }
   }
   if (!cdptr) cdptr = ContentSectionPtr;

   if (cdptr->TitlePtr)
      TitlePtr = cdptr->TitlePtr;
   else
      TitlePtr = cdptr->SectionTitle;

   if (cptr = TitlePtr)
   {
      /* copy transmogrifying back- to forward- slashes */
      zptr = (sptr = TitleBuffer) + sizeof(TitleBuffer)-1;
      /* if it comes from 'cdptr->TitlePtr' then it will be '$$'-delimited */
      while (*cptr && *(USHORTPTR)cptr != '$$' && sptr < zptr)
      {
         if (*cptr == '\\')
            *sptr++ = '/'; 
         else
            *sptr++ = *cptr;
         cptr++;
      } 
      *sptr = '\0';
      if (!*(cptr = TitleBuffer)) cptr = NULL;
   }

   if (!cptr)
   {
      if (ContentSectionPtr)
         cptr = ContentSectionPtr->SectionTitle;
      else
         cptr = ContentSectionName;

      /* copy transmogrifying back- to forward- slashes */
      zptr = (sptr = TitleBuffer) + sizeof(TitleBuffer)-1;
      while (*cptr && sptr < zptr)
      {
         if (*cptr == '\\')
            *sptr++ = '/'; 
         else
            *sptr++ = *cptr;
         cptr++;
      } 
      *sptr = '\0';
      cptr = TitleBuffer;
   }

   fputs (HtmlEscape(cptr), stdout);
}

/*****************************************************************************/
/*
*/

void TemplateTop (char* SubsArg)

{
   char  *cptr, *sptr, *zptr,
         *TopPtr;
   char  TopString [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateTop() |%s|\n", SubsArg);

   if (*(cptr = SubsArg) == '$') cptr++;
   zptr = (sptr = TopString) + sizeof(TopString)-1;
   while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';
   if (!*(TopPtr = TopString))
      TopPtr = CONFIG_LANG_PARSE(LangNavigate,4);

   if (cptr = IsAnImage(TopPtr))
      fprintf (stdout,
"<a href=\"#TOP\"><img class=\"vwcms_nav_img\" alt=\"[TOP]\" %s /></a>",
               cptr);
   else
      fprintf (stdout,
"<a href=\"#TOP\"><span class=\"vwcms_nav\">%s</span></a>",
               TopPtr);
}

/*****************************************************************************/
/*
If a user is logged-in then output the name.  Two optional arguments can be
provided, the first is prepended to the name, the second appended.
*/

void TemplateUser (char *SubsArg)

{
   char  *cptr, *sptr, *zptr;
   char  StrBuf [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateUser() |%s|\n", SubsArg);

   if (!LoginUserName[0]) return;

   if (*SubsArg == '$') SubsArg++; else SubsArg = "";

   cptr = SubsArg;
   zptr = (sptr = StrBuf) + sizeof(StrBuf)-1;
   while (*cptr && *cptr != '$' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';
   if (StrBuf[0]) fputs (StrBuf, stdout);

   fputs (HtmlEscape(LoginUserName), stdout);

   if (*cptr) cptr++;
   zptr = (sptr = StrBuf) + sizeof(StrBuf)-1;
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';
   if (StrBuf[0]) fputs (StrBuf, stdout);
}

/*****************************************************************************/
/*
Build a null-terminated string containing the directive and any parameters in
dynamic memory.  Return a pointer to the next character in the content.
Removes the leading and trailing double-dollars.
*/

char* TemplateParseDirect
(
DYSTRING *dyptr,
char *cptr
)
{
   char  *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateParseDirect()\n");

   if (*(USHORTPTR)cptr != '$$') ErrorExit (SS$_BUGCHECK, FI_LI);
   cptr += 2;

   dyptr->blen = 0;
   if (!dyptr->csize) dyptr->csize = 1024;
   zptr = (sptr = dyptr->bptr) + dyptr->bsize;
   for (;;)
   {
      if (sptr >= zptr)
      {
         dyptr->blen = sptr - dyptr->bptr;
         dyptr->bsize += dyptr->csize;
         if (dyptr->vmem)
            dyptr->bptr = CgiLibVeeMemRealloc (dyptr->bptr, dyptr->bsize+32);
         else
            dyptr->bptr = realloc (dyptr->bptr, dyptr->bsize+32);
         if (!dyptr->bptr) ErrorExit (vaxc$errno, FI_LI);
         zptr = (sptr = dyptr->bptr) + dyptr->bsize;
         sptr += dyptr->blen;
      }
      if (!*cptr || *(USHORTPTR)cptr == '$$') break;
      if (*(USHORTPTR)cptr == '\\$') *sptr++ = *cptr++;
      *sptr++ = *cptr++;
   }
   *sptr = '\0';
   dyptr->blen = sptr - dyptr->bptr;
   if (Debug) fprintf (stdout, "|%s|\n", dyptr->bptr);

   if (*(USHORTPTR)cptr == '$$') cptr += 2;
   return (cptr);
}

/*****************************************************************************/
/*
Find a directive ("$$BLAH$") in 'raw' content.  Return a pointer to the first
character or to the terminating null if not found.  An immediate parameter can
be appended to the directive ("$$BLAH$param=").
*/

char* TemplateParseFind
(
char *cptr,
char *sptr
)
{
   int  len;
   char  *aptr;
   char  toups [4];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateParseFind() |%s|\n", sptr);

   /* supplied directive must begin with '$$' and sort-of end with '$' */
   if (*(USHORTPTR)sptr != '$$') ErrorExit (SS$_BUGCHECK, FI_LI);
   for (aptr = sptr + 2; *aptr != '$'; aptr++);
   if (*aptr != '$') ErrorExit (SS$_BUGCHECK, FI_LI);
   /* allow for finding an immediate parameter as well */
   while (*aptr) aptr++;
   len = aptr - sptr;
   if (len < 4) ErrorExit (SS$_BUGCHECK, FI_LI);

   /* bit of a search optimisation */
   toups[2] = toupper(sptr[2]);
   toups[3] = toupper(sptr[3]);
   sptr += 4;
   len -= 4;

   while (*cptr)
   {
      while (*cptr && *(USHORTPTR)cptr != '$$') cptr++;
      if (!*cptr) break;
      if (*(ULONGPTR)cptr == '$$$$')
         cptr += 4;
      else
      if (toupper(cptr[2]) == toups[2] &&
          toupper(cptr[3]) == toups[3] &&
          strsame (cptr+4, sptr, len))
      {
         if (Debug) fprintf (stdout, "|%s|\n", cptr);
         return (cptr);
      }
      else
      if (isalpha(cptr[2]))
      {
         /* scan to the end of this directive */
         for (cptr += 2; *cptr && *(USHORTPTR)cptr != '$$'; cptr++)
            if (*(USHORTPTR)cptr == '\\$') cptr++;
         if (!*cptr) break;
         cptr += 2;
      }
      else
         cptr += 2;
   }
   if (Debug) fprintf (stdout, "|%s|\n", cptr);
   return (cptr);
}

/*****************************************************************************/
/*
Find a directive parameter in 'raw' content.  Return a pointer to the first
character of the supplied parameter name or to the directive terminating '$$'
if not found.
*/

char* TemplateParseFindParam
(
char *cptr,
char *sptr
)
{
   int  len;
   char  *aptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateParseFindParam() |%s|\n", sptr);

   for (aptr = sptr; *aptr != '='; aptr++);
   len = aptr - sptr;
   if (*aptr == '=') len++;

   while (*cptr && *(USHORTPTR)cptr != '$$')
   {
      while (*cptr && *cptr == '$') cptr++;
      if (strsame (cptr, sptr, len))
         if (*aptr == '=' || cptr[len] == '$')
            return (cptr);
      while (*cptr && *cptr != '$') cptr++;
   }
   return (cptr);
}

/*****************************************************************************/
/*
Parse a dollar-delimitted string.  If the first character is a '=' then this is
stepped over.  Return a pointer to next character (if any).  Can parse into a
null-sptr when it just discards.  A null character or a double-dollar (for
'raw' content parsing) can terminate the processing.  If a double-dollar the
pointer returns the first of those two characters.
*/

char* TemplateParseParam
(
char *cptr,
char *sptr,
int sizes
)
{
   char  *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateParseParam() |%32.32s|\n", cptr);

   if (!cptr)
      cptr = "";
   else
   if (*cptr == '=')
      cptr++;

   if (sptr)
   {
      if (sizes > 0) sizes--; else sizes = 0;
      zptr = sptr + sizes;
      while (*cptr && *cptr != '$' && sptr < zptr)
      {
         if (*(USHORTPTR)cptr == '\\$') cptr++;
         *sptr++ = *cptr++;
      }
      *sptr = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", sptr-(sizes-(int)(zptr-sptr)));
   }
   /* in case of buffer overflow */
   while (*cptr && *cptr != '$')
   {
      if (*(USHORTPTR)cptr == '\\$') cptr++;
      cptr++;
   }
   /* step over any next dollar delimitter */
   if (*cptr && *(USHORTPTR)cptr != '$$') cptr++;
   return (cptr);
}

/*****************************************************************************/
/*
As with TemplateParseParam() but just skips to the next dollar.
*/

char* TemplateParseSkipParam
(
char *cptr,
char *sptr
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateParseSkipParam() |%s|\n", cptr);

   if (sptr) *sptr = '\0';
   while (*cptr && *cptr != '$')
   {
      if (*(USHORTPTR)cptr == '\\$') cptr++;
      cptr++;
   }
   if (*(USHORTPTR)cptr == '\\$') cptr++;
   if (*cptr && *(USHORTPTR)cptr != '$$') cptr++;
   return (cptr);
}

/*****************************************************************************/
/*
As with TemplateParseParam() but parses one or two intgers from the text.
*/

char* TemplateParseRowCol
(
char *cptr,
int *rowptr,
int *colptr
)
{
   int  rows, cols;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TemplateParseRowCol() |%s|\n", cptr);

   rows = atoi(cptr);
   if (rowptr) *rowptr = rows;
   while (isdigit(*cptr)) cptr++;
   while (*cptr && *cptr != '$' && !isdigit(*cptr)) cptr++;
   cols = atoi(cptr);
   if (colptr) *colptr = cols;
   while (*cptr && *cptr != '$')
   {
      if (*(USHORTPTR)cptr == '\\$') cptr++;
      cptr++;
   }
   if (*(USHORTPTR)cptr == '\\$') cptr++;
   if (*cptr && *(USHORTPTR)cptr != '$$') cptr++;
   return (cptr);
}

/*****************************************************************************/
/*
Display the error highlighted in-line to the content.
*/

void TemplateInlineError
(
char *DirectPtr,
char *SubsArgPtr
)
{
   /*********/
   /* begin */
   /*********/

   if (SubsArgPtr)
   {
      if (*SubsArgPtr == '$') SubsArgPtr++;
      fprintf (stdout, "<span class=\"vwcms_error\">$$%s$%s$$</span>",
               DirectPtr, SubsArgPtr);
   }
   else
      fprintf (stdout, "<span class=\"vwcms_error\">$$%s$$</span>", DirectPtr);
}

/*****************************************************************************/
/*
Report substitution parameter problem.
*/

void TemplateParamError
(
char *DirectPtr,
char *SubsArgPtr,
char *ParamPtr
)
{
   /*********/
   /* begin */
   /*********/

   if (SubsArgPtr)
      TemplateInlineError (DirectPtr, HtmlEscape(SubsArgPtr));
   else
      TemplateInlineError (DirectPtr, NULL);

   if (ParamPtr)
      TemplateAdvise ("E\'!AZ\' !AZ", HtmlEscape(ParamPtr),
                      CONFIG_LANG_PARSE(LangSubsError,2));
   else
      TemplateAdvise ("E!AZ", CONFIG_LANG_PARSE(LangSubsError,2));
}

/*****************************************************************************/
/*
Report some problem or another.
*/

void TemplateReportError
(
char *DirectPtr,
char *SubsArgPtr,
char *ReportPtr
)
{
   /*********/
   /* begin */
   /*********/

   if (SubsArgPtr)
      TemplateInlineError (DirectPtr, HtmlEscape(SubsArgPtr));
   else
      TemplateInlineError (DirectPtr, NULL);

   if (ReportPtr)
      TemplateAdvise ("E$$!AZ$$ !AZ", DirectPtr, HtmlEscape(ReportPtr));
   else
      TemplateAdvise ("E$$!AZ$$ !AZ", DirectPtr, "?");
}

/*****************************************************************************/

